from __future__ import print_function

import numpy as np


class StatusLine:
    def __init__(self, video):
        """A status line from an image frame is read.  The last line of the
        Photonfocus camra picture is the status line.  All data is returned in
        the form of a struct, as well as the cut-frame without the status line.

        Parameters
        ----------
        video: NDArray
            Numpy array with the following form should be used for video
            (number of frame, width, height).

        """

        self.video = video
        self.out_video = None
        self.status_line_row_flag = False
        self.status_line_column_flag = False
        self.camera_info = {}

    def find_status_line(self):
        """
        Returns
        -------
        self.out_video: NDArray
            Video without status line.

        self.camera_info: dic
            The dictionary that illustrates the obtain information in the status line.

        """

        if np.all(self.video[0, -1, :4] == [0xFF, 0x00, 0xAA, 0x55]):
            self.status_line_column_flag = True
            self.read_status_line(self.video[0])
            print("---Status line detected in column---")
            self.camera_info["status_line_position"] = "column"
            return self.out_video, self.camera_info

        elif np.all(self.video[0, :4, -1] == [0xFF, 0x00, 0xAA, 0x55]):
            self.status_line_row_flag = True
            self.read_status_line(self.video[0])
            print("---Status line detected in row---")
            self.camera_info["status_line_position"] = "row"
            return self.out_video, self.camera_info

        else:
            print("---Status line does not detect---")
            self.camera_info["status_line_position"] = ""
            return self.video, self.camera_info

    def dec2hex(self, n):
        return "%X" % int(n)

    def hex2dec(self, s):
        return int(s, 16)

    def decimalToBinary(self, num, len_):
        d2b_ = np.zeros(len_)
        d2b = bin(int(num)).replace("0b", "")
        tmp_ = [int(d_) for d_ in d2b]
        diff_len = abs(len_ - len(tmp_))
        d2b_[diff_len:] = tmp_

        return d2b_

    def read_status_line(self, frame):
        """
        Parameters
        ----------
        frame: NDArray
            First frame in video.
        """

        # constants for our camera
        clock_cycle = 80e6  # Camera Clock Cycle in Hz
        nr_taps = 2  # Nr of Taps
        if self.status_line_column_flag:
            cutframe = frame[:, 0:-1]
            self.out_video = self.video[:, 1:-1, :]

            width = frame.shape[0]
            converter = [1, 256, 256 * 256, 256 * 256 * 256]
            StatusLine = frame[-1, :]

        elif self.status_line_row_flag:
            cutframe = frame[0:-1, :]
            self.out_video = self.video[:, :, 1:-1]

            width = frame.shape[1]
            converter_ = [1, 256, 256 * 256, 256 * 256 * 256]
            converter = np.asarray(converter_).astype(np.int64)
            StatusLine = frame[:, -1].astype(np.int64)

        firstbits = sum(np.multiply(StatusLine[0:4], converter))
        NrAcc = firstbits / self.hex2dec("55AA00FF")

        if (NrAcc % 1) != 0:
            print("\n... It looks like there is no Status line!?")

        self.NrAcc = NrAcc
        self.Preamble = self.dec2hex(firstbits / NrAcc)  # Preamble (0x55AA00FF)

        if width >= 8:
            ImgCnt = sum(np.multiply(StatusLine[4:8], converter)) / NrAcc  # (Mean) Image Counter
            self.camera_info["ImgCnt"] = ImgCnt

        if width >= 12:
            RealTimeCnt = (
                sum(np.multiply(StatusLine[8:12], converter)) / NrAcc
            )  # (Mean) Real Time Counter in Microseconds
            self.camera_info["RealTimeCnt"] = RealTimeCnt

        if width >= 16:
            MissedTrgCnt = (
                sum(np.multiply(StatusLine[12:16], converter)) / NrAcc
            )  # (Mean) Missed Trigger Counter
            self.camera_info["MissedTrgCnt"] = MissedTrgCnt

        if width >= 20:
            self.ImgAvgVal = (
                sum(np.multiply(StatusLine[16:20], converter)) / NrAcc
            )  # (Mean) Image Average Value
            self.camera_info["RealTimeCnt"] = RealTimeCnt

        if width >= 24:
            IntTimeClk = (
                sum(np.multiply(StatusLine[20:24], converter)) / NrAcc
            )  # Integration Time in units of clock cycles (80MHz)
            IntTime = IntTimeClk / clock_cycle * nr_taps * 1000  # Integration Time in ms
            self.camera_info["IntTimeClk"] = IntTimeClk
            self.camera_info["IntTime"] = IntTime

        if width >= 28:
            BrstTrgNr = (
                sum(np.multiply(StatusLine[24:28], converter)) / NrAcc
            )  # (Mean) Burst Trigger Number
            self.camera_info["BrstTrgNr"] = BrstTrgNr

        if width >= 32:
            MissedBrstTrgCnt = (
                sum(np.multiply(StatusLine[28:32], converter)) / NrAcc
            )  # (Mean) Missed Burst Trigger Counter
            self.camera_info["MissedBrstTrgCnt"] = MissedBrstTrgCnt

        if width >= 36:
            ROIx_start = (
                sum(np.multiply(StatusLine[32:36], converter)) / NrAcc
            )  # Horizontal start position of ROI (Window.X)
            self.camera_info["ROIx_start"] = ROIx_start

        if width >= 40:
            ROIx_end = (
                sum(np.multiply(StatusLine[36:40], converter)) / NrAcc
            )  # Horizontal end position of ROI (= Window.X + Window.W - 1)
            NrCols = ROIx_end - ROIx_start + 1
            self.camera_info["ROIx_end"] = ROIx_end
            self.camera_info["NrCols"] = NrCols

        if width >= 44:
            ROIy_start = (
                sum(np.multiply(StatusLine[40:44], converter)) / NrAcc
            )  # Vertical start position of ROI (Window.Y). In MROI-mode this
            # parameter is the start position of the first ROI
            self.camera_info["ROIy_start"] = ROIy_start

        if width >= 48:
            NrRows = sum(np.multiply(StatusLine[44:48], converter)) / NrAcc + 1  # Number of rows
            ROIy_end = ROIy_start + NrRows - 1
            self.camera_info["NrRows"] = NrRows
            self.camera_info["ROIy_end"] = ROIy_end

        if width >= 52:
            TrgSrc = (
                sum(np.multiply(StatusLine[48:52], converter)) / NrAcc
            )  # Trigger Source (0=Free Running, 1=Interface Trigger, 2=I/O Trigger)
            self.camera_info["TrgSrc"] = TrgSrc

        if width >= 56:
            DigGain = 2 ** (
                sum(np.multiply(StatusLine[52:56], converter)) / NrAcc
            )  # Digital Gain (1x, 2x, 4x, 8x)
            self.camera_info["DigGain"] = DigGain

        if width >= 60:
            DigOffset = sum(np.multiply(StatusLine[56:60], converter)) / NrAcc  # Digital Offset
            self.camera_info["DigOffset"] = DigOffset

        if width >= 64:
            CamType = (
                sum(np.multiply(StatusLine[60:64], converter)) / NrAcc
            )  # Camera Type Code (Our Photonfocus "MV1-D1024E-160-CL-12" is 110)
            self.camera_info["CamType"] = CamType

        if width >= 68:
            SerialNr = (
                sum(np.multiply(StatusLine[64:68], converter)) / NrAcc
            )  # Camera Serial Number
            self.camera_info["SerialNr"] = SerialNr

        if width >= 80:
            tmp_ = list(range(3, -13, -1))
            tmp_0 = np.float_power(2, tmp_)
            tmp_1 = np.multiply(StatusLine[76:80], converter)
            tmp_2 = int(np.sum(tmp_1) / self.NrAcc)
            tmp_3 = self.decimalToBinary(tmp_2, 16)
            tmp_4 = np.multiply(tmp_3, tmp_0)
            FineGain = np.sum(tmp_4)
            self.camera_info["FineGain"] = FineGain

        if width >= 96:
            TrgLvl = self.decimalToBinary(
                int(np.sum(np.multiply(StatusLine[92:96], converter))) / NrAcc, 4
            )  # Trigger Level: signals level of the trigger input signals. Bit
            # 0: ExSync (CC1): Bit 1: I/O Trigger; Bit 2: CC3; Bit 3: CC4
            self.camera_info["TrgLvl"] = TrgLvl



import numpy as np
from ipywidgets import Layout, interact, widgets
from matplotlib import cm as cm
from matplotlib import pyplot as plt
from matplotlib.patches import Circle, Rectangle
from mpl_toolkits.axes_grid1 import make_axes_locatable
from scipy.ndimage import median_filter

from piscat.InputOutput import read_status_line
from piscat.Localization import gaussian_2D_fit
from piscat.Trajectory.data_handling import protein_trajectories_list2dic


class JupyterDisplay:
    def __init__(
        self,
        video,
        median_filter_flag=False,
        color="gray",
        title=None,
        xlabel=None,
        ylabel=None,
        imgSizex=5,
        imgSizey=5,
        extent=None,
        IntSlider_width="500px",
        step=1,
    ):
        """This class displays the video in jupyter notebook.

        Parameters
        ----------
        video:(NDArray)
            Input video.

        median_filter_flag: bool
            In case it defines as True, a median filter is applied with size 3
            to remove hot pixel effect.

        color: str
            It defines the colormap for visualization.

        title: list
            A list of string titles with a length equal to the number of frames
            in the video.

        xlabel: str
            The label text for x-axis.

        ylabel: str
            The label text for y-axis.

        imgSizex: int
            Image length size.

        imgSizey: int
            Image width size.

        extent:
            The extent kwarg determines the bounding box in data coordinates
            that the image will fill, which is specified in data coordinates as
            (left, right, bottom, top).

        IntSlider_width: str
            Size of slider

        step: int
            Stride between visualization frames.

        """
        self.color = color
        self.video = video
        self.imgSizex = imgSizex
        self.imgSizey = imgSizey
        self.title = title
        self.extent = extent

        self.median_filter_flag = median_filter_flag

        self.xlabel = xlabel
        self.ylabel = ylabel

        interact(
            self.display,
            frame=widgets.IntSlider(
                min=0,
                max=self.video.shape[0] - 1,
                step=step,
                value=10,
                layout=Layout(width=IntSlider_width),
                readout_format="10",
                continuous_update=False,
                description="Frame:",
            ),
        )

    def display(self, frame):
        fig = plt.figure(figsize=(self.imgSizex, self.imgSizey))
        ax = fig.add_axes([0, 0, 1, 1])

        if self.median_filter_flag:
            frame_v = median_filter(self.video[int(frame), :, :], 3)
        else:
            frame_v = self.video[int(frame), :, :]

        if self.extent is not None:
            myplot = ax.imshow(frame_v, cmap=self.color, extent=self.extent)
        else:
            myplot = ax.imshow(frame_v, cmap=self.color)

        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        plt.colorbar(myplot, cax=cax)

        if self.xlabel is not None:
            ax.set_xlabel(self.xlabel)

        if self.ylabel is not None:
            ax.set_ylabel(self.ylabel)

        if self.title is not None:
            ax.set_title(self.title[int(frame)])

        plt.show()


class JupyterDisplay_StatusLine:
    def __init__(
        self,
        video,
        median_filter_flag=False,
        color="gray",
        imgSizex=5,
        imgSizey=5,
        IntSlider_width="500px",
        step=1,
        value=0,
    ):
        """This class displays the video in the Jupyter notebook interactively
        while highlight status line

        Parameters
        ----------
        video: NDArray
            Input video.

        median_filter_flag: bool
            In case it defines as True, a median filter is applied with size 3
            to remove hot pixel effect.

        color: str
            It defines the colormap for visualization.

        imgSizex: int
            Image length size.

        imgSizey: int
            Image width size.

        IntSlider_width: str
            Size of slider.

        step: int
            Stride between visualization frames.

        value: int
            Initial frame value for visualization

        """
        self.color = color
        self.video = video
        self.imgSizex = imgSizex
        self.imgSizey = imgSizey
        self.median_filter_flag = median_filter_flag

        status_ = read_status_line.StatusLine(video)
        self.video_remove, status_information = status_.find_status_line()
        self.statusLine_position = status_information["status_line_position"]

        interact(
            self.display,
            frame=widgets.IntSlider(
                min=0,
                max=self.video.shape[0] - 1,
                step=step,
                value=value,
                layout=Layout(width=IntSlider_width),
                readout_format="10",
                continuous_update=False,
                description="Frame:",
            ),
        )

    def display(self, frame):
        fig = plt.figure(figsize=(self.imgSizex, self.imgSizey))
        ax = fig.add_axes([0, 0, 1, 1])

        if self.median_filter_flag:
            frame_v = median_filter(self.video[int(frame), :, :], 3)
            frame_v_rm = median_filter(self.video_remove[int(frame), :, :], 3)

        else:
            frame_v = self.video[int(frame), :, :]
            frame_v_rm = self.video_remove[int(frame), :, :]

        myplot_rm = ax.imshow(frame_v_rm, cmap=self.color)
        myplot = ax.imshow(frame_v, cmap=self.color)
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        plt.colorbar(myplot_rm, cax=cax)

        if self.statusLine_position == "column":
            rect = Rectangle(
                (0, self.video.shape[2]),
                self.video.shape[2],
                1,
                linewidth=10,
                edgecolor="r",
                facecolor="none",
            )
            ax.add_patch(rect)
        elif self.statusLine_position == "row":
            rect = Rectangle(
                (self.video.shape[1], 0),
                1,
                self.video.shape[1],
                linewidth=10,
                edgecolor="r",
                facecolor="none",
            )
            ax.add_patch(rect)
        plt.show()


class JupyterPSFs_localizationDisplay:
    def __init__(
        self,
        video,
        df_PSFs,
        median_filter_flag=False,
        color="gray",
        imgSizex=5,
        imgSizey=5,
        IntSlider_width="500px",
        step=1,
        value=0,
    ):
        """This class displays the video in the Jupyter notebook interactively
        while highlight PSFs.

        Parameters
        ----------
        video: NDArray
           Input video.

        df_PSFs: panda data_frame
            Data Frames that contains the location of PSFs.

        median_filter_flag: bool
           In case it defines as True, a median filter is applied with size 3
           to remove hot pixel effect.

        color: str
           It defines the colormap for visualization.

        imgSizex: int
           Image length size.

        imgSizey: int
           Image width size.

        IntSlider_width: str
           Size of slider.

        step: int
           Stride between visualization frames.

        value: int
           Initial frame value for visualization

        """
        self.video = video
        self.imgSizex = imgSizex
        self.imgSizey = imgSizey
        self.df_PSFs = df_PSFs
        self.median_filter_flag = median_filter_flag
        self.color = color

        interact(
            self.show_psf,
            frame_number=widgets.IntSlider(
                min=0,
                max=self.video.shape[0] - 1,
                step=step,
                value=value,
                readout_format="1",
                continuous_update=False,
                layout=Layout(width=IntSlider_width),
                description="Frame:",
            ),
        )

    def show_psf(self, frame_number):
        fig = plt.figure(figsize=(self.imgSizex, self.imgSizey))
        ax = fig.add_axes([0, 0, 1, 1])
        if self.median_filter_flag:
            frame_v = median_filter(self.video[int(frame_number), :, :], 3)
        else:
            frame_v = self.video[int(frame_number), :, :]

        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        myplot = ax.imshow(frame_v, cmap=self.color)
        plt.colorbar(myplot, cax=cax)

        particle = self.df_PSFs.loc[self.df_PSFs["frame"] == frame_number]
        particle_X = particle["x"].tolist()
        particle_Y = particle["y"].tolist()
        particle_sigma = particle["sigma"].tolist()

        for j_ in range(len(particle_X)):
            y = int(particle_Y[j_])
            x = int(particle_X[j_])
            sigma = particle_sigma[j_]
            ax.add_patch(
                Circle(
                    (x, y),
                    radius=np.sqrt(2) * sigma,
                    edgecolor="r",
                    facecolor="none",
                    linewidth=2,
                )
            )
        plt.show()


class JupyterPSFs_localizationPreviewDisplay:
    def __init__(
        self,
        video,
        df_PSFs,
        title="",
        frame_num=None,
        median_filter_flag=False,
        color="gray",
        imgSizex=5,
        imgSizey=5,
        IntSlider_width="500px",
        step=1,
    ):
        """This class displays the video in the Jupyter notebook interactively
        while highlight PSFs.

        Parameters
        ----------
        video: NDArray
           Input video.

        df_PSFs: panda data_frame
            Data Frames that contains the location of PSFs.

        title: str
            It defines title of plot.

        frame_num: list
            list of frame that we want to see preview of localization.

        median_filter_flag: bool
           In case it defines as True, a median filter is applied with size 3
           to remove hot pixel effect.

        color: str
           It defines the colormap for visualization.

        imgSizex: int
           Image length size.

        imgSizey: int
           Image width size.

        IntSlider_width: str
           Size of slider.

        step: int
           Stride between visualization frames.

        """
        self.video = video
        self.imgSizex = imgSizex
        self.imgSizey = imgSizey
        self.df_PSFs = df_PSFs
        self.median_filter_flag = median_filter_flag
        self.color = color
        if self.df_PSFs is not None:
            self.frame_num = np.sort(np.unique(df_PSFs["frame"].tolist()))
        else:
            self.frame_num = frame_num
        self.currentFrame = 0
        self.IntSlider_width = IntSlider_width
        self.step = step
        self.title = title

    def display_run(self):
        interact(
            self.show_psf,
            index=widgets.IntSlider(
                min=0,
                max=len(self.frame_num) - 1,
                step=self.step,
                value=self.currentFrame,
                readout_format="1",
                continuous_update=False,
                layout=Layout(width=self.IntSlider_width),
                description="index:",
            ),
        )

    def show_psf(self, index):
        frame_number = self.frame_num[index]
        self.currentFrame = frame_number
        fig = plt.figure(figsize=(self.imgSizex, self.imgSizey))
        ax = fig.add_axes([0, 0, 1, 1])
        if self.median_filter_flag:
            frame_v = median_filter(self.video[int(frame_number), :, :], 3)
        else:
            frame_v = self.video[int(frame_number), :, :]

        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        myplot = ax.imshow(frame_v, cmap=self.color)
        ax.set_title(self.title)
        plt.colorbar(myplot, cax=cax)

        if self.df_PSFs is not None:
            particle = self.df_PSFs.loc[self.df_PSFs["frame"] == frame_number]
            particle_X = particle["x"].tolist()
            particle_Y = particle["y"].tolist()
            particle_sigma = particle["sigma"].tolist()

            for j_ in range(len(particle_X)):
                y = int(particle_Y[j_])
                x = int(particle_X[j_])
                sigma = particle_sigma[j_]
                ax.add_patch(
                    Circle(
                        (x, y),
                        radius=np.sqrt(2) * sigma,
                        edgecolor="r",
                        facecolor="none",
                        linewidth=2,
                    )
                )
        plt.show()

    def set_df(self, new_df):
        self.df_PSFs = new_df


class JupyterPSFs_subplotLocalizationDisplay:
    def __init__(
        self,
        list_videos,
        list_df_PSFs,
        numRows,
        numColumns,
        list_titles=None,
        median_filter_flag=False,
        color="gray",
        imgSizex=5,
        imgSizey=5,
        IntSlider_width="500px",
        step=1,
        value=0,
    ):
        """This class shows several videos (with the same number of frames) at
        once in the Jupyter notebook interactively while highlight localize
        PSFs.

        Parameters
        ----------
        list_videos: list of NDArray
          List of videos.

        list_df_PSFs: list panda data_frame
           List Data Frames that contains the location of PSFs for each video.

        numRows: int
            It defines number of rows in sub-display.

        numColumns: int
            It defines number of columns in sub-display.

        list_titles: list str
            List of titles for each sub plot.

        median_filter_flag: bool
          In case it defines as True, a median filter is applied with size 3 to
          remove hot pixel effect.

        color: str
          It defines the colormap for visualization.

        imgSizex: int
          Image length size.

        imgSizey: int
          Image width size.

        IntSlider_width: str
          Size of slider

        step: int
          Stride between visualization frames.

        value: int
            Initial frame value for visualization

        """
        self.list_video = list_videos
        self.numVideos = len(list_videos)
        self.imgSizex = imgSizex
        self.imgSizey = imgSizey
        self.list_df_PSFs = list_df_PSFs
        self.median_filter_flag = median_filter_flag
        self.color = color
        self.numRows = numRows
        self.numColumns = numColumns

        if list_titles is None:
            self.list_titles = [None for _ in range(self.numVideos)]
        else:
            self.list_titles = list_titles

        max_numberFrames = np.max([vid_.shape[0] for vid_ in list_videos])
        interact(
            self.show_psf,
            frame_number=widgets.IntSlider(
                min=0,
                max=max_numberFrames,
                step=step,
                value=value,
                readout_format="1",
                continuous_update=False,
                layout=Layout(width=IntSlider_width),
                description="Frame:",
            ),
        )

    def show_psf(self, frame_number):
        fig = plt.figure(figsize=(self.imgSizex, self.imgSizey))
        grid = plt.GridSpec(self.numRows, self.numColumns, hspace=0.3, wspace=0.3)

        imgGrid_list = []
        for i in range(self.numRows):
            for j in range(self.numColumns):
                imgGrid_list.append(fig.add_subplot(grid[i, j]))

        for img_, tit_, img_grid_, df_PSFs in zip(
            self.list_video, self.list_titles, imgGrid_list, self.list_df_PSFs
        ):
            if self.median_filter_flag:
                frame_v = median_filter(img_[int(frame_number), :, :], 3)
            else:
                frame_v = img_[int(frame_number), :, :]

            divider = make_axes_locatable(img_grid_)
            cax = divider.append_axes("right", size="5%", pad=0.05)
            myplot = img_grid_.imshow(frame_v, cmap=self.color)
            img_grid_.axis("off")
            plt.colorbar(myplot, cax=cax)
            if tit_ is not None:
                img_grid_.set_title(tit_)

            particle = df_PSFs.loc[df_PSFs["frame"] == frame_number]
            particle_X = particle["x"].tolist()
            particle_Y = particle["y"].tolist()
            particle_sigma = particle["sigma"].tolist()

            for j_ in range(len(particle_X)):
                y = int(particle_Y[j_])
                x = int(particle_X[j_])
                sigma = particle_sigma[j_]
                img_grid_.add_patch(
                    Circle(
                        (x, y),
                        radius=np.sqrt(2) * sigma,
                        edgecolor="r",
                        facecolor="none",
                        linewidth=2,
                    )
                )
        plt.show()


class JupyterPSFs_TrackingDisplay:
    def __init__(
        self,
        video,
        df_PSFs,
        median_filter_flag=False,
        step=1,
        color="gray",
        imgSizex=5,
        imgSizey=5,
        value=0,
    ):
        """This class displays video in the Jupyter notebook interactively
        while highlighting PSFs with trajectories.

        Parameters
        ----------
        video: NDArray
           Input video.

        df_PSFs: panda data_frame
            Data Frames that contains the location of PSFs.

        median_filter_flag: bool
           In case it defines as True, a median filter is applied with size 3
           to remove hot pixel effect.

        color: str
           It defines the colormap for visualization.

        imgSizex: int
           Image length size.

        imgSizey: int
           Image width size.

        IntSlider_width: str
           Size of slider.

        step: int
           Stride between visualization frames.

        value: int
            Initial frame value for visualization

        """

        self.video = video
        self.df_PSFs = df_PSFs
        self.imgSizex = imgSizex
        self.imgSizey = imgSizey
        self.df_PSFs = df_PSFs
        self.median_filter_flag = median_filter_flag
        self.color = color
        self.median_filter_flag = median_filter_flag
        if self.df_PSFs.particle.isnull().any():
            self.df_PSFs["particle"] = 0
            self.list_particles_idx = self.df_PSFs.particle.unique()
        else:
            self.list_particles_idx = self.df_PSFs.particle.unique()

        colors_ = cm.autumn(np.linspace(0, 1, len(self.list_particles_idx)))

        self.colors = colors_[0 : len(self.list_particles_idx), :]
        interact(
            self.show_psf,
            frame_number=widgets.IntSlider(
                min=0,
                max=self.video.shape[0] - 1,
                step=step,
                value=value,
                readout_format="1",
                continuous_update=False,
                description="Frame:",
            ),
        )

    def show_psf(self, frame_number):
        fig = plt.figure(figsize=(self.imgSizex, self.imgSizey))
        ax = fig.add_axes([0, 0, 1, 1])

        if self.median_filter_flag:
            frame_v = median_filter(self.video[int(frame_number), :, :], 3)
        else:
            frame_v = self.video[int(frame_number), :, :]

        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        myplot = ax.imshow(frame_v, cmap=self.color)
        plt.colorbar(myplot, cax=cax)

        particle = self.df_PSFs.loc[self.df_PSFs["frame"] == frame_number]
        particle_X = particle["x"].tolist()
        particle_Y = particle["y"].tolist()
        particle_sigma = particle["sigma"].tolist()
        particle_labels = particle["particle"].tolist()

        if len(particle_labels) != 0:
            for j_, p_l in zip(range(len(particle_X)), particle_labels):
                y = int(particle_Y[j_])
                x = int(particle_X[j_])
                sigma = particle_sigma[j_]
                ax = plt.gca()
                ax.add_patch(
                    Circle(
                        (x, y),
                        radius=np.sqrt(2) * sigma,
                        edgecolor=self.colors[int(p_l)],
                        facecolor="none",
                        linewidth=2,
                    )
                )

                all_particle_ = self.df_PSFs.loc[self.df_PSFs["particle"] == p_l]
                particle_f_ = all_particle_["frame"].tolist()
                particle_X_ = all_particle_["x"].tolist()
                particle_Y_ = all_particle_["y"].tolist()
                particle_sigma_ = all_particle_["sigma"].tolist()
                for f_, x_, y_, _ in zip(particle_f_, particle_X_, particle_Y_, particle_sigma_):
                    if f_ <= frame_number:
                        ax.add_patch(
                            Circle(
                                (int(x_), int(y_)),
                                radius=0.1,
                                edgecolor=self.colors[int(p_l)],
                                facecolor="none",
                                linewidth=1,
                            )
                        )
                        # ax.add_collection((x_, y_), autolim=True, cmap=self.colors [p_l])
        plt.show()


class JupyterSelectedPSFs_localizationDisplay:
    def __init__(
        self,
        video,
        particles,
        particles_num="#0",
        frame_extend=0,
        median_filter_flag=False,
        flag_fit2D=False,
        color="gray",
        imgSizex=10,
        imgSizey=10,
        IntSlider_width="500px",
        step=1,
        value=0,
    ):
        """This class interactively shows video in a Jupyter notebook while
        highlighting PSFs based on ID.

        Parameters
        ----------
        video: NDArray
           Input video.

        particles: dic
             Dictionary similar to the following structures.:

            | {"#0": {'intensity_horizontal': ..., 'intensity_vertical': ...,
               ..., 'particle_ID': ...}, "#1": {}, ...}

        particles_num: str
            Choose the corresponding key in the particles dictionary.

        frame_extend: int
            Display particle for ``frame_extend`` before and after segmented
            ones. In case there are not enough frames before/after, it shows
            only for the number of existing frames.

        flag_fit2D: bool
            It activate 2D-Gaussian fit to extract fitting information of selected PSF.

        median_filter_flag: bool
           In case it defines as True, a median filter is applied with size 3
           to remove hot pixel effect.

        color: str
           It defines the colormap for visualization.

        imgSizex: int
           Image length size.

        imgSizey: int
           Image width size.

        IntSlider_width: str
           Size of slider.

        step: int
           Stride between visualization frames.

        value: int
            Initial frame value for visualization

        """

        if type(particles) is list:
            particles = protein_trajectories_list2dic(particles)

        if type(particles) is dict:
            for key, item in particles.items():
                if key == particles_num:
                    intensity_horizontal = item["intensity_horizontal"]
                    intensity_vertical = item["intensity_vertical"]
                    center_int = item["center_int"]
                    center_int_flow = item["center_int_flow"]
                    self.frame_number_ = item["frame_number"]
                    self.sigma = item["sigma"]
                    self.x_center = item["x_center"]
                    self.y_center = item["y_center"]
                    self.particle_ID = item["particle_ID"]

        self.imgSizex = imgSizex
        self.imgSizey = imgSizey
        self.video = video
        self.color = color
        self.flag_fit2D = flag_fit2D
        self.median_filter_flag = median_filter_flag

        video_frameNum = video.shape[0]
        max_particle_frame = np.max(self.frame_number_)
        min_particle_frame = np.min(self.frame_number_)

        max_extend_particle_frame = np.min([max_particle_frame + frame_extend, video_frameNum])
        min_extend_particle_frame = np.max([min_particle_frame - frame_extend, 0])

        self.frame_number = list(range(min_extend_particle_frame, max_extend_particle_frame, 1))
        interact(
            self.show_psf,
            f_num=widgets.IntSlider(
                min=self.frame_number[0],
                max=self.frame_number[-1] - 1,
                step=step,
                value=value,
                readout_format="1",
                continuous_update=False,
                layout=Layout(width=IntSlider_width),
                description="Frame:",
            ),
        )

    def show_psf(self, f_num):
        fig = plt.figure(figsize=(self.imgSizex, self.imgSizey))
        ax = fig.add_axes([0, 0, 1, 1])
        if self.flag_fit2D:
            fit_params = self.fit_2D_gussian(frame_num=f_num, scale=5)

        if self.median_filter_flag:
            frame_v = median_filter(self.video[int(f_num), :, :], 3)
        else:
            frame_v = self.video[int(f_num), :, :]

        myplot = ax.imshow(frame_v, cmap=self.color)
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        plt.colorbar(myplot, cax=cax)

        if f_num <= np.max(self.frame_number_) and f_num >= np.min(self.frame_number_):
            idx_ = np.where(self.frame_number_ == f_num)

            particle_X = self.x_center[idx_]
            particle_Y = self.y_center[idx_]
            particle_sigma = self.sigma[idx_]
            particle_labels = self.particle_ID[idx_]

            y = int(particle_Y)
            x = int(particle_X)
            sigma = particle_sigma
            ax.add_patch(
                Circle(
                    (x, y),
                    radius=np.sqrt(2) * sigma,
                    edgecolor="r",
                    facecolor="none",
                    linewidth=2,
                )
            )

        plt.show()

    def fit_2D_gussian(self, frame_num, scale=5):
        particle = self.list_psf.loc[self.list_psf["frame"] == frame_num]
        particle_X = particle["x"].tolist()
        particle_Y = particle["y"].tolist()
        particle_sigma = particle["sigma"].tolist()
        particle_center_intensity = particle["center_intensity"].tolist()

        index_list = [index for index in particle.index]

        list_one_frame_fit = []
        if particle.shape[0] > 0:
            for p_x, p_y, sigma_0, _, i_ in zip(
                particle_X, particle_Y, particle_sigma, particle_center_intensity, index_list
            ):
                window_size = scale * np.sqrt(2) * sigma_0
                start_sigma = sigma_0
                if p_x > window_size and p_y > window_size:
                    window_frame = self.video[
                        int(frame_num),
                        int(p_y - window_size) + 1 : int(p_y + window_size),
                        int(p_x - window_size) + 1 : int(p_x + window_size),
                    ]
                    w_s = window_size
                else:
                    window_size_temp = window_size
                    while p_x < window_size_temp or p_y < window_size_temp:
                        window_size_temp = window_size_temp - 2
                    window_frame = self.video[
                        int(frame_num),
                        int(p_y - window_size_temp) + 1 : int(p_y + window_size_temp),
                        int(p_x - window_size_temp) + 1 : int(p_x + window_size_temp),
                    ]
                    w_s = window_size_temp

                fit_params_ = gaussian_2D_fit.fit_2D_Gaussian_varAmp(
                    window_frame, sigma_x=start_sigma, sigma_y=start_sigma, display_flag=False
                )
                fit_params = fit_params_[1]
                fit_errors = fit_params_[2]

                print("Fit Amplitude:", fit_params[0], "\u00b1", fit_errors[0])
                print("Fit X-Center: ", fit_params[1], "\u00b1", fit_errors[1])
                print("Fit Y-Center: ", fit_params[2], "\u00b1", fit_errors[2])
                print("Fit X-Sigma:  ", fit_params[3], "\u00b1", fit_errors[3])
                print("Fit Y-Sigma:  ", fit_params[4], "\u00b1", fit_errors[4])
                print("Fit Bias:  ", fit_params[5], "\u00b1", fit_errors[5])

                fit_params = [frame_num, i_, fit_params_, p_x, p_y, w_s]
                return fit_params


class JupyterSubplotDisplay:
    def __init__(
        self,
        list_videos,
        numRows,
        numColumns,
        list_titles=None,
        imgSizex=20,
        imgSizey=20,
        IntSlider_width="500px",
        median_filter_flag=False,
        color="gray",
        step=1,
        value=0,
        vmin=None,
        vmax=None,
    ):
        """This class interactively displays several videos (with the same
        number of frames) in a Jupyter notebook.

        Parameters
        ----------
        list_videos: list of NDArray
          List of videos

        numRows: int
            It defines number of rows in sub-display

        numColumns: int
            It defines number of columns in sub-display

        list_titles: list str
            List of titles for each sub plot

        median_filter_flag: bool
          In case it defines as True, a median filter is applied with size 3 to
          remove hot pixel effect.

        color: str
            It defines the colormap for visualization.

        imgSizex: int
            Image length size.

        imgSizey: int
            Image width size.

        IntSlider_width: str
            Size of slider

        step: int
            Stride between visualization frames.

        value: int
            Initial frame value for visualization

        """

        self.color = color
        self.vmin = vmin
        self.vmax = vmax

        self.video = list_videos
        self.median_filter_flag = median_filter_flag
        self.numRows = numRows
        self.numColumns = numColumns
        self.numVideos = len(list_videos)
        self.imgSizex = imgSizex
        self.imgSizey = imgSizey

        if list_titles is None:
            self.list_titles = [None for _ in range(self.numVideos)]
        else:
            self.list_titles = list_titles

        max_numberFrames = np.max([vid_.shape[0] for vid_ in list_videos])

        interact(
            self.display,
            frame=widgets.IntSlider(
                min=0,
                max=max_numberFrames - 1,
                step=step,
                value=value,
                layout=Layout(width=IntSlider_width),
                readout_format="100",
                continuous_update=False,
                description="Frame:",
            ),
        )

    def display(self, frame):
        fig = plt.figure(figsize=(self.imgSizex, self.imgSizey))

        grid = plt.GridSpec(self.numRows, self.numColumns, hspace=0.3, wspace=0.3)

        imgGrid_list = []
        for i in range(self.numRows):
            for j in range(self.numColumns):
                imgGrid_list.append(fig.add_subplot(grid[i, j]))

        for img_, tit_, img_grid_ in zip(self.video, self.list_titles, imgGrid_list):
            if self.median_filter_flag:
                frame_v = median_filter(img_[int(frame), :, :], 3)
            else:
                frame_v = img_[int(frame), :, :]

            # img_grid_.imshow(frame_v, cmap=self.color)
            # img_grid_.axis('off')

            divider = make_axes_locatable(img_grid_)
            cax = divider.append_axes("right", size="5%", pad=0.05)
            if self.vmin is None or self.vmax is None:
                myplot = img_grid_.imshow(frame_v, cmap=self.color)
            else:
                myplot = img_grid_.imshow(
                    frame_v, cmap=self.color, vmin=self.vmin, vmax=self.vmax
                )
            img_grid_.axis("off")
            plt.colorbar(myplot, cax=cax)
            if tit_ is not None:
                img_grid_.set_title(tit_)

        plt.show()


class JupyterPSFs_2_modality_subplotLocalizationDisplay:
    def __init__(
        self,
        list_videos,
        list_df_PSFs_1,
        list_df_PSFs_2,
        numRows,
        numColumns,
        list_titles=None,
        median_filter_flag=False,
        color="gray",
        imgSizex=5,
        imgSizey=5,
        IntSlider_width="500px",
        step=1,
        value=0,
        edgecolor_1="r",
        edgecolor_2="g",
    ):
        """This class will interactively sub-display multiple videos in a
        Jupyter notebook while highlighting PSFs determined in two modalities.

        Parameters
        ----------
        list_videos: list of NDArray
          List of videos.

        list_df_PSFs_1: list
            list of all data Frames that contains the location of PSFs in first modality.

        list_df_PSFs_2: list
            list of all data Frames that contains the location of PSFs in second modality.

        numRows: int
            It defines number of rows in sub-display.

        numColumns: int
            It defines number of columns in sub-display.

        list_titles: list str
            List of titles for each sub plot.

        median_filter_flag: bool
          In case it defines as True, a median filter is applied with size 3 to
          remove hot pixel effect.

        color: str
          It defines the colormap for visualization.

        imgSizex: int
          Image length size.

        imgSizey: int
          Image width size.

        IntSlider_width: str
          Size of slider

        step: int
          Stride between visualization frames.

        value: int
            Initial frame value for visualization

        edgecolor_1: str
            The color of the circle that annotates the first modality on the display.

        edgecolor_2: str
            The color of the circle that annotates the second modality on the display.

        """
        self.list_video = list_videos
        self.numVideos = len(list_videos)
        self.imgSizex = imgSizex
        self.imgSizey = imgSizey
        self.list_df_PSFs_1 = list_df_PSFs_1
        self.list_df_PSFs_2 = list_df_PSFs_2
        self.edgecolor_1 = edgecolor_1
        self.edgecolor_2 = edgecolor_2
        self.median_filter_flag = median_filter_flag
        self.color = color
        self.numRows = numRows
        self.numColumns = numColumns

        if list_titles is None:
            self.list_titles = [None for _ in range(self.numVideos)]
        else:
            self.list_titles = list_titles

        max_numberFrames = np.max([vid_.shape[0] for vid_ in list_videos])
        interact(
            self.show_psf,
            frame_number=widgets.IntSlider(
                min=0,
                max=max_numberFrames - 1,
                step=step,
                value=value,
                readout_format="1",
                continuous_update=False,
                layout=Layout(width=IntSlider_width),
                description="Frame:",
            ),
        )

    def show_psf(self, frame_number):
        fig = plt.figure(figsize=(self.imgSizex, self.imgSizey))
        grid = plt.GridSpec(self.numRows, self.numColumns, hspace=0.3, wspace=0.3)

        imgGrid_list = []
        for i in range(self.numRows):
            for j in range(self.numColumns):
                imgGrid_list.append(fig.add_subplot(grid[i, j]))

        for img_, tit_, img_grid_, df_PSFs_1, df_PSFs_2 in zip(
            self.list_video,
            self.list_titles,
            imgGrid_list,
            self.list_df_PSFs_1,
            self.list_df_PSFs_2,
        ):
            if self.median_filter_flag:
                frame_v = median_filter(img_[int(frame_number), :, :], 3)
            else:
                frame_v = img_[int(frame_number), :, :]

            divider = make_axes_locatable(img_grid_)
            cax = divider.append_axes("right", size="5%", pad=0.05)
            myplot = img_grid_.imshow(frame_v, cmap=self.color)
            img_grid_.axis("off")
            plt.colorbar(myplot, cax=cax)
            if tit_ is not None:
                img_grid_.set_title(tit_)

            particle_1 = df_PSFs_1.loc[df_PSFs_1["frame"] == frame_number]
            particle_X_1 = particle_1["x"].tolist()
            particle_Y_1 = particle_1["y"].tolist()
            particle_sigma_1 = particle_1["sigma"].tolist()

            particle_2 = df_PSFs_2.loc[df_PSFs_2["frame"] == frame_number]
            particle_X_2 = particle_2["x"].tolist()
            particle_Y_2 = particle_2["y"].tolist()
            particle_sigma_2 = particle_2["sigma"].tolist()

            for j_ in range(len(particle_X_1)):
                y = int(particle_Y_1[j_])
                x = int(particle_X_1[j_])
                sigma = particle_sigma_1[j_]
                img_grid_.add_patch(
                    Circle(
                        (x, y),
                        radius=np.sqrt(2) * sigma,
                        edgecolor=self.edgecolor_1,
                        facecolor="none",
                        linewidth=2,
                    )
                )

            for j_ in range(len(particle_X_2)):
                y = int(particle_Y_2[j_])
                x = int(particle_X_2[j_])
                sigma = particle_sigma_2[j_]
                img_grid_.add_patch(
                    Circle(
                        (x, y),
                        radius=np.sqrt(2) * sigma,
                        edgecolor=self.edgecolor_2,
                        facecolor="none",
                        linewidth=2,
                    )
                )

        plt.show()


class JupyterFPNcDisplay:
    def __init__(
        self,
        list_videos,
        list_titles=None,
        correction_axis=0,
        numRows=1,
        numColumns=2,
        imgSizex=20,
        imgSizey=20,
        IntSlider_width="500px",
        median_filter_flag=False,
        color="gray",
        step=1,
        value=0,
    ):
        """This class can sub-display several FPNc video in the Jupyter
        notebook interactively while 1D projection on the direction of
        correction axis illustrates below on each subplot.

        Parameters
        ----------
        list_videos: list of NDArray
          List of videos

        correction_axis: (0/1)
            selecting the axis that FPNc correction should visualize for it.

        numRows: int
            It defines number of rows in sub-display.

        numColumns: int
            It defines number of columns in sub-display.

        list_titles: list str
            List of titles for each sub plot.

        median_filter_flag: bool
            In case it defines as True, a median filter is applied with size 3
            to remove hot pixel effect.

        color: str
            It defines the colormap for visualization.

        imgSizex: int
            Image length size.

        imgSizey: int
            Image width size.

        IntSlider_width: str
            Size of slider

        step: int
            Stride between visualization frames.

        value: int
            Initial frame value for visualization

        """
        self.color = color
        self.video = list_videos
        self.median_filter_flag = median_filter_flag
        self.numRows = numRows
        self.numColumns = numColumns
        self.numVideos = len(list_videos)
        self.imgSizex = imgSizex
        self.imgSizey = imgSizey
        self.correction_axis = correction_axis

        if list_titles is None:
            self.list_titles = [None for _ in range(self.numVideos)]
        else:
            self.list_titles = list_titles

        max_numberFrames = np.max([vid_.shape[0] for vid_ in list_videos])

        interact(
            self.display,
            frame=widgets.IntSlider(
                min=0,
                max=max_numberFrames,
                step=step,
                value=value,
                layout=Layout(width=IntSlider_width),
                readout_format="100",
                continuous_update=False,
                description="Frame:",
            ),
        )

    def display(self, frame):
        self.fig = plt.figure(figsize=(self.imgSizex, self.imgSizey))

        grid = self.fig.add_gridspec(
            nrows=self.numRows, ncols=self.numColumns, left=0.05, right=0.48, wspace=0.05
        )
        axs = grid.subplots(sharex="col", sharey="row")

        for img_, tit_, img_grid_ in zip(self.video, self.list_titles, axs):
            if self.median_filter_flag:
                frame_v = median_filter(img_[int(frame), :, :], 3)
            else:
                frame_v = img_[int(frame), :, :]
            if self.correction_axis == 0:
                img_grid_.imshow(frame_v, cmap=self.color)
            elif self.correction_axis == 1:
                img_grid_.imshow(np.transpose(frame_v), cmap=self.color)
            if tit_ is not None:
                img_grid_.set_title(tit_)

        for ax in axs.flat:
            ax.label_outer()

        plt.show()

        fig = plt.figure(figsize=(self.imgSizex, self.imgSizey / 10))
        grid = fig.add_gridspec(
            nrows=self.numRows, ncols=self.numColumns, left=0.05, right=0.48, wspace=0.05
        )
        axs = grid.subplots(sharex="col", sharey="row")

        for img_, img_grid_ in zip(self.video, axs):
            frame_v = img_[int(frame), :, :]
            gainMap1D_median = np.mean(frame_v, axis=self.correction_axis)

            x_grid = list(range(0, frame_v.shape[1]))
            y_grid = list(range(0, frame_v.shape[0]))

            x, y = np.meshgrid(x_grid, y_grid)
            if self.correction_axis == 0:
                img_grid_.plot(
                    np.ravel(x), np.ravel(frame_v, order="A"), "b.", linewidth=1, markersize=0.5
                )
            elif self.correction_axis == 1:
                img_grid_.plot(
                    np.ravel(x), np.ravel(frame_v, order="F"), "b.", linewidth=1, markersize=0.5
                )

            img_grid_.plot(gainMap1D_median, "r-", linewidth=1, markersize=0.5)
            min_y = frame_v.mean() - 4 * frame_v.std()
            max_y = frame_v.mean() + 4 * frame_v.std()
            img_grid_.set_ylim(min_y, max_y)
            img_grid_.ticklabel_format(style="sci", axis="y", scilimits=(0, 0))

        # Hide x labels and tick labels for all but bottom plot.
        for ax in axs.flat:
            ax.label_outer()

        plt.show()



from PySide6 import QtCore, QtWidgets

from piscat.GUI.CPU_Configurations import Import_configuration
from piscat.InputOutput.cpu_configurations import CPUConfigurations


class CPU_setting_wrapper(QtWidgets.QMainWindow):
    def __init__(self):
        super(CPU_setting_wrapper, self).__init__()
        self.change_cpu_setting()

    def change_cpu_setting(self):
        print("\nChange CPU setting-->", end="")
        self.cpu_setting = Import_configuration.CPU_Setting()
        self.cpu_setting.update_CPU_Setting.connect(self.save_CPU_setting)

    @QtCore.Slot()
    def save_CPU_setting(self):
        if self.cpu_setting.parallel_active is True:
            setting_dic = {
                "n_jobs": [self.cpu_setting.n_jobs],
                "backend": [self.cpu_setting.backend],
                "verbose": [self.cpu_setting.verbose],
                "parallel_active": [self.cpu_setting.parallel_active],
                "threshold_for_parallel_run": [self.cpu_setting.threshold_for_parallel_run],
            }

            cpu_setting = CPUConfigurations()
            cpu_setting.save_cpu_setting(setting_dic)
        else:
            setting_dic = {
                "n_jobs": [1],
                "backend": ["None"],
                "verbose": [0],
                "parallel_active": [self.cpu_setting.parallel_active],
                "threshold_for_parallel_run": [None],
            }
            cpu_setting = CPUConfigurations()
            cpu_setting.save_cpu_setting(setting_dic)
        print("Done!")



from PySide6 import QtCore, QtWidgets


class DRA(QtWidgets.QMainWindow):
    signal_finish = QtCore.Signal(bool)

    def __init__(self, parent=None):
        super().__init__(parent)
        self.batch_size = None
        self.axis = 1
        self.mode_FPN = "cpFPN"

        self.flag_display = False
        self.flag_FPN = False
        self.flag_power_normalization = False

        self.window = QtWidgets.QWidget()

        self.le1 = QtWidgets.QLineEdit()
        self.le1.setPlaceholderText("radius")

        self.ok = QtWidgets.QPushButton("Ok")
        self.ok.setAutoDefault(False)
        self.ok.clicked.connect(self.do_update)

        self.checkbox_display = QtWidgets.QCheckBox("Display", self)
        self.checkbox_FPN = QtWidgets.QCheckBox("Fix pattern noise", self)
        self.checkbox_power_normalization = QtWidgets.QCheckBox("Laser power normalization", self)

        self.grid = QtWidgets.QGridLayout()
        self.grid.addWidget(self.le1, 0, 0)
        self.grid.addWidget(self.checkbox_display, 0, 1)
        self.grid.addWidget(self.checkbox_power_normalization, 1, 1)
        self.grid.addWidget(self.createFirstExclusiveGroup(), 1, 0)
        self.grid.addWidget(self.ok, 2, 0)

        self.window.setWindowTitle("Rolling Average")
        self.setStyleSheet("QMainWindow{background-color: darkgray;}")
        self.window.setLayout(self.grid)
        self.window.show()

    def createFirstExclusiveGroup(self):
        self.groupBox_FPNc = QtWidgets.QGroupBox("FPNc:")
        self.groupBox_FPNc.setCheckable(True)
        self.groupBox_FPNc.setChecked(False)

        self.FPN_mode_group = QtWidgets.QButtonGroup()
        self.radio_wFPN_mode = QtWidgets.QRadioButton("Wavelet FPNc")
        self.radio_cpFPN_mode = QtWidgets.QRadioButton("Column_projection FPNc")
        self.radio_wf_FPN_mode = QtWidgets.QRadioButton("Wavelet_FFT2D FPNc")
        self.radio_cpFPN_mode.setChecked(True)

        self.FPN_mode_group.addButton(self.radio_wFPN_mode)
        self.FPN_mode_group.addButton(self.radio_cpFPN_mode)
        self.FPN_mode_group.addButton(self.radio_wf_FPN_mode)

        self.axis_group = QtWidgets.QButtonGroup()
        self.radio_axis_1 = QtWidgets.QRadioButton("FPNc in axis 0")
        self.radio_axis_2 = QtWidgets.QRadioButton("FPNc in axis 1")
        self.radio_axis_3 = QtWidgets.QRadioButton("FPNc in Both axis")
        self.radio_axis_2.setChecked(True)

        self.axis_group.addButton(self.radio_axis_1)
        self.axis_group.addButton(self.radio_axis_2)
        self.axis_group.addButton(self.radio_axis_3)

        grid = QtWidgets.QGridLayout()
        grid.addWidget(self.radio_cpFPN_mode, 0, 0)
        grid.addWidget(self.radio_wFPN_mode, 0, 1)
        grid.addWidget(self.radio_wf_FPN_mode, 0, 2)

        grid.addWidget(self.radio_axis_1, 1, 0)
        grid.addWidget(self.radio_axis_2, 1, 1)
        grid.addWidget(self.radio_axis_3, 1, 2)

        self.groupBox_FPNc.setLayout(grid)
        return self.groupBox_FPNc

    def do_update(self):
        if self.ok.clicked:
            self.get_values()
            if self.checkbox_display.isChecked():
                self.flag_display = True
            else:
                self.flag_display = False

            if self.groupBox_FPNc.isChecked():
                self.flag_FPN = True
                if self.radio_axis_1.isChecked():
                    self.axis = 0
                elif self.radio_axis_2.isChecked():
                    self.axis = 1
                elif self.radio_axis_3.isChecked():
                    self.axis = "Both"

                if self.radio_cpFPN_mode.isChecked():
                    self.mode_FPN = "cpFPN"
                elif self.radio_wFPN_mode.isChecked():
                    self.mode_FPN = "wFPN"
                elif self.radio_wf_FPN_mode.isChecked():
                    self.mode_FPN = "fFPN"

            if self.checkbox_power_normalization.isChecked():
                self.flag_power_normalization = True

            if self.batch_size != "":
                self.batch_size = int(self.batch_size)
                self.signal_finish.emit(True)

            else:
                self.msg_box = QtWidgets.QMessageBox()
                self.msg_box.setWindowTitle("Warning!")
                self.msg_box.setText("Please insert batch size!")
                self.msg_box.exec_()

    def get_values(self):
        self.batch_size = self.le1.text()

    def closeEvent(self, event):
        QtCore.QCoreApplication.instance().quit()



from .data_handling import *
from .directional_intensity import *
from .gaussian_2D_fit import *
from .localization_filtering import *
from .particle_localization import PSFsExtraction as PSFsExtraction
from .radial_symmetry_centering import *



from __future__ import print_function

import cv2
import numpy as np
import scipy.fftpack
import scipy.ndimage
import scipy.signal
from joblib import Parallel, delayed
from PySide6.QtCore import QObject, QRunnable, Signal, Slot
from scipy.ndimage import median_filter
from skimage import filters
from tqdm.autonotebook import tqdm

from piscat.InputOutput.cpu_configurations import CPUConfigurations


class WorkerSignals(QObject):
    finished = Signal()
    error = Signal(tuple)
    result = Signal(object)
    progress = Signal(int)


class Filters:
    def __init__(self, video, inter_flag_parallel_active=True):
        """This class generates a list of video/image filters.  To improve
        performance on large video files, some of them have a parallel
        implementation.

        Parameters
        ----------
        video: NDArray
            The video is 3D-numpy (number of frames, width, height).

        inter_flag_parallel_active: bool
            If the user wants to enable general parallel tasks in the CPU
            configuration, he or she can only use this flag to enable or
            disable this process.

        """
        self.cpu = CPUConfigurations()
        self.inter_flag_parallel_active = inter_flag_parallel_active

        self.video = video
        self.filtered_video = None

    def temporal_median(self):
        """By extracting the temporal median from pixels, the background is
        corrected.

        Returns
        -------
        @returns: NDArray
            The background corrected video as 3D-numpy
        """
        video_med = np.median(self.video, axis=0)
        video_med_ = np.expand_dims(video_med, axis=0)
        return np.divide(self.video, video_med_) - 1

    def flat_field(self, sigma):
        """This method corrects the video background by creating a synthetic
        flat fielding version of the background.

        Parameters
        ----------
        sigma: float
            Sigma of Gaussian filter that use to create blur video.

        Returns
        -------
        flat_field_video: NDArray
            The background corrected video as 3D-numpy

        """
        blur_video = self.gaussian(sigma)
        flat_field_video = np.divide(self.video, blur_video) - 1
        return flat_field_video

    def median(self, size):
        """
        This function applies a 2D median filter on each frame.

        Parameters
        ----------
        size: int
            Kernel size of the median filter.

        Returns
        -------
        self.blur_video: NDArray
            The filter video as 3D-numpy.

        """
        if self.cpu.parallel_active is True and self.inter_flag_parallel_active is True:
            print("\n---start median filter with Parallel---")

            result = Parallel(
                n_jobs=self.cpu.n_jobs, backend=self.cpu.backend, verbose=self.cpu.verbose
            )(delayed(self.median_kernel)(x, size) for x in tqdm(range(self.video.shape[0])))

            arry_result = np.asarray(result)
            self.filtered_video = np.reshape(
                arry_result, (len(result), self.video.shape[1], self.video.shape[2])
            )
        else:
            print("\n---start median filter without Parallel---")

            result = [
                median_filter(self.video[i_, :, :], size) for i_ in range(self.video.shape[0])
            ]
            arry_result = np.asarray(result)
            self.filtered_video = np.reshape(
                arry_result, (len(result), self.video.shape[1], self.video.shape[2])
            )
        return self.filtered_video

    def median_kernel(self, i_, size):
        return median_filter(self.video[i_, :, :], size)

    def gaussian(self, sigma):
        """
        This function applies a 2D Gaussian filter on each frame.

        Parameters
        ----------
        sigma: float
            Sigma of Gaussian filter that use to create blur video.

        Returns
        -------
        self.blur_video: NDArray
            The filter video as 3D-numpy.

        """

        if self.cpu.parallel_active is True and self.inter_flag_parallel_active is True:
            print("\n---start gaussian filter with Parallel---")

            result = Parallel(
                n_jobs=self.cpu.n_jobs, backend=self.cpu.backend, verbose=self.cpu.verbose
            )(delayed(self.gaussian_kernel)(x, sigma) for x in tqdm(range(self.video.shape[0])))

            arry_result = np.asarray(result)
            self.filtered_video = np.reshape(
                arry_result, (len(result), self.video.shape[1], self.video.shape[2])
            )
        else:
            print("\n---start gaussian filter without Parallel---")

            result = [
                filters.gaussian(self.video[i_, :, :], sigma=sigma)
                for i_ in range(self.video.shape[0])
            ]
            arry_result = np.asarray(result)
            self.filtered_video = np.reshape(
                arry_result, (len(result), self.video.shape[1], self.video.shape[2])
            )

        return self.filtered_video

    def gaussian_kernel(self, i_, sigma):
        return filters.gaussian(self.video[i_, :, :], sigma=sigma, preserve_range=True)


class FFT2D(QRunnable):
    def __init__(self, video):
        """
        This class computes the 2D spectrum of video.

        Parameters
        ----------
        video: NDArray
            The video is 3D-numpy (number of frames, width, height).
        """
        super(FFT2D, self).__init__()
        self.signals = WorkerSignals()
        self.video = video

    @Slot()
    def run(self):
        result_ = self.fft2D()
        result = self.log2_scale(result_)
        self.signals.result.emit(result)

    def fft2D(self):
        im_fft = np.ndarray(self.video.shape, complex)
        for i in range(0, self.video.shape[0]):
            im_fft[i, :, :] = np.fft.fftshift(np.fft.fft2(self.video[i, :, :]))
        return im_fft

    def log2_scale(self, video):
        return np.log2(np.abs(video))


class RadialVarianceTransform:
    def __init__(self, inter_flag_parallel_active=True):
        """Efficient Python implementation of Radial Variance Transform.

        The main function is :func:`rvt` in the bottom of the file, which
        applies the transform to a single image (2D numpy array).


        Compared to the vanilla convolution implementation, there are two speed-ups:

        1) Pre-calculating and caching kernel FFT; this way so only one inverse
        FFT is calculated per convolution + one direct fft of the image is used
        for all convolutions

        2) When finding MoV, calculate ``np.mean(rsqmeans)`` in a single
        convolution by averaging all kernels first

        Parameters
        ----------
        inter_flag_parallel_active: bool
            In case the user wants to active general parallel tasks in CPU configuration,
            the user can only active or deactivate this method by this flag.

        References
        ----------
        The Radial Variance Transform code has been adopted from the GitHub repository mentioned below.

            [1] Kashkanova, Anna D., et al. "Precision single-particle localization
            using radial variance transform." Optics Express 29.7 (2021):
            11070-11083.

            [2] https://github.com/SandoghdarLab/rvt

        """
        self.cpu = CPUConfigurations()
        self._kernels_fft_cache = {}
        self.video = None
        self.inter_flag_parallel_active = inter_flag_parallel_active

    def gen_r_kernel(self, r, rmax):
        """Generate a ring kernel with radius `r` and size ``2*rmax+1``"""
        a = rmax * 2 + 1
        k = np.zeros((a, a))
        for i in range(a):
            for j in range(a):
                rij = ((i - rmax) ** 2.0 + (j - rmax) ** 2) ** 0.5
                if int(rij) == r:
                    k[i][j] = 1.0

        tmp = k / np.sum(k)
        return tmp

    def generate_all_kernels(self, rmin, rmax, coarse_factor=1, coarse_mode="add"):
        """Generate a set of kernels with radii between `rmin` and `rmax` and
        sizes ``2*rmax+1``.

        ``coarse_factor`` and ``coarse_mode`` determine if the number of those
        kernels is reduced by either skipping or adding them (see :func:`rvt`
        for a more detail explanation).

        """
        kernels = [self.gen_r_kernel(r, rmax) for r in range(rmin, rmax + 1)]
        if coarse_factor > 1:
            if coarse_mode == "skip":
                kernels = kernels[::coarse_factor]
            else:
                kernels = [
                    np.sum(kernels[i : i + coarse_factor], axis=0) / coarse_factor
                    for i in range(0, len(kernels), coarse_factor)
                ]
        return kernels

    def _check_core_args(self, rmin, rmax, kind, coarse_mode="add"):
        """Check validity of the core algorithm arguments"""
        if rmin < 0 or rmax < 0:
            raise ValueError("radius should be non-negative")
        if rmin >= rmax:
            raise ValueError("rmax should be strictly greater than rmin")
        if kind not in {"basic", "normalized"}:
            raise ValueError("unrecognized kind: {}; can be either 'basic' or 'normalized'")
        if coarse_mode not in {"add", "skip"}:
            raise ValueError("unrecognized coarse mode: {}; can be either 'add' or 'skip'")

    def _check_args(self, rmin, rmax, kind, coarse_mode, highpass_size, upsample):
        """Check validity of all the algorithm arguments"""
        self._check_core_args(rmin, rmax, kind, coarse_mode)
        if upsample < 1:
            raise ValueError("upsampling factor should be positive")
        if highpass_size is not None and highpass_size < 0.3:
            raise ValueError("high-pass filter size should be >= 0.3")

    ## Prepare auxiliary parameters
    def get_fshape(self, s1, s2, fast_mode=False):
        """Get the required shape of the transformed image given the shape of
        the original image and the kernel."""
        shape = s1 if fast_mode else s1 + s2 - 1
        fshape = [scipy.fftpack.next_fast_len(int(d)) for d in shape]

        tmp = tuple(fshape)
        return tmp

    ## Preparing FFTs of arrays
    def prepare_fft(self, inp, fshape, pad_mode="constant"):
        """Prepare the image for a convolution by taking its Fourier transform,
        applying padding if necessary."""
        if pad_mode == "fast":
            return np.fft.rfftn(inp, fshape)
        else:
            pad = [((td - d) // 2, (td - d + 1) // 2) for td, d in zip(fshape, inp.shape)]
            inp = np.pad(inp, pad, mode=pad_mode)
            tmp = np.fft.rfftn(inp)
            return tmp

    ## Shortcut of SciPy fftconvolve, which takes already fft'd arrays on the input
    def convolve_fft(self, sp1, sp2, s1, s2, fshape, fast_mode=False):
        """Calculate the convolution from the Fourier transforms of the
        original image and the kernel, trimming the result if necessary."""
        ret = np.fft.irfftn(sp1 * sp2, fshape)
        if fast_mode:
            return np.roll(ret, (-(s2[0] // 2), -(s2[1] // 2)), (0, 1))[: s1[0], : s1[1]].copy()
        else:
            off = (fshape[0] - s1[0]) // 2 + s2[0] // 2, (fshape[1] - s1[1]) // 2 + s2[1] // 2
            tmp = ret[off[0] : off[0] + s1[0], off[1] : off[1] + s1[1]].copy()
            return tmp

    def rvt_core(
        self,
        img,
        rmin,
        rmax,
        kind="basic",
        rweights=None,
        coarse_factor=1,
        coarse_mode="add",
        pad_mode="constant",
    ):
        """Perform core part of Radial Variance Transform (RVT) of an image.

        Parameters
        ----------
        img: NDArray
            source image (2D numpy array)

        rmin: float
            minimal radius (inclusive)

        rmax: float
            maximal radius (inclusive)

        kind: str, ("basic", "normalized")
            either ``"basic"`` (only VoM), or ``"normalized"`` (VoM/MoV);
            normalized version increases subpixel bias, but it works better at
            lower SNR

        rweights:
            relative weights of different radial kernels; must be a 1D array of
            the length ``(rmax-rmin+1)//coarse_factor``

        coarse_factor:
            the reduction factor for the number ring kernels; can be used to
            speed up calculations at the expense of precision

        coarse_mode: str, ("add", "skip")
            The reduction method; can be ``"add"`` (add ``coarse_factor`` rings
            in a row to get a thicker ring, which works better for smooth
            features), or ``"skip"`` (only keep on in ``coarse_factor`` rings,
            which works better for very fine features)

        pad_mode: str, ("constant", "reflect", "edge", "fast")
            Edge padding mode for convolutions; can be either one of modes
            accepted by ``np.pad`` (such as ``"constant"``, ``"reflect"``, or
            ``"edge"``), or ``"fast"``, which means faster no-padding (a
            combination of ``"wrap"`` and ``"constant"`` padding depending on
            the image size); ``"fast"`` mode works faster for smaller images
            and larger ``rmax``, but the border pixels (within ``rmax`` from
            the edge) are less reliable; note that the image mean is subtracted
            before processing, so ``pad_mode="constant"`` (default) is
            equivalent to padding with a constant value equal to the image mean

        """
        self._check_core_args(rmin, rmax, kind, coarse_mode)  # check arguments validity
        s1 = np.array(img.shape)
        s2 = np.array([rmax * 2 + 1, rmax * 2 + 1])
        fast_mode = pad_mode == "fast"
        fshape = self.get_fshape(
            s1, s2, fast_mode=fast_mode
        )  # calculate the padded image shape (add padding and get to the next "good" FFT size)
        cache_k = (rmin, rmax, coarse_factor, coarse_mode) + fshape
        if (
            cache_k not in self._kernels_fft_cache
        ):  # generate convolution kernels, if they are not in cache yet
            kernels = self.generate_all_kernels(
                rmin, rmax, coarse_factor=coarse_factor, coarse_mode=coarse_mode
            )
            self._kernels_fft_cache[cache_k] = [
                self.prepare_fft(k, fshape, pad_mode="fast") for k in kernels
            ]  # pad_mode="fast" corresponds to the default zero-padding here
        kernels_fft = self._kernels_fft_cache[
            cache_k
        ]  # get the convolution kernels (either newely generated, or cached)
        if rweights is not None:
            rweights = np.asarray(rweights)
            if len(rweights) != len(kernels_fft):
                raise ValueError(
                    "The number of weights {} differs from the number of kernels {}".format(
                        len(rweights), len(kernels_fft)
                    )
                )
            rweights = rweights / rweights.sum()  # normalize weights by their sum
        img = (
            img - img.mean()
        )  # subtract mean (makes VoM calculation more stable and zero-padding more meaningful)
        img_fft = self.prepare_fft(
            img, fshape, pad_mode=pad_mode
        )  # prepare image FFT (only needs to be done once)
        rmeans = np.array(
            [
                self.convolve_fft(img_fft, k_fft, s1, s2, fshape, fast_mode=fast_mode)
                for k_fft in kernels_fft
            ]
        )  # calculate M_r for all radii
        if rweights is None:
            vom = np.var(
                rmeans, axis=0
            )  # calculate VoM as a standard variance of M_r along the radius axis
        else:
            vom = (
                np.sum(rmeans**2 * rweights[:, None, None], axis=0)
                - np.sum(rmeans * rweights[:, None, None], axis=0) ** 2
            )  # calculate VoM as a weighted variance of M_r along the radius axis
        if kind == "basic":
            return vom
        else:  # calculate MoV for normalization
            imgsq_fft = self.prepare_fft(img**2, fshape, pad_mode=pad_mode)  # prepare image FFT
            if rweights is None:
                sumk_fft = np.mean(kernels_fft, axis=0)  # find combined kernel as a standard mean
                mov = self.convolve_fft(
                    imgsq_fft, sumk_fft, s1, s2, fshape, fast_mode=fast_mode
                ) - np.mean(
                    rmeans**2, axis=0
                )  # use the combined kernel to find MoV in one convolution
            else:
                sumk_fft = np.sum(
                    kernels_fft * rweights[:, None, None], axis=0
                )  # find combined kernel as a weighted mean
                mov = self.convolve_fft(
                    imgsq_fft, sumk_fft, s1, s2, fshape, fast_mode=fast_mode
                ) - np.sum(
                    rmeans**2 * rweights[:, None, None], axis=0
                )  # use the combined kernel to find MoV in one convolution
            tmp = vom / mov
            return tmp

    def high_pass(self, img, size):
        """Perform Gaussian high-pass filter on the image"""
        img = img.astype(float)
        tmp = img - scipy.ndimage.gaussian_filter(img, size)
        return tmp

    def rvt(
        self,
        img,
        rmin,
        rmax,
        kind="basic",
        highpass_size=None,
        upsample=1,
        rweights=None,
        coarse_factor=1,
        coarse_mode="add",
        pad_mode="constant",
    ):
        """Perform Radial Variance Transform (RVT) of an image.

        Parameters
        ----------
        img: NDArray
            source image (2D numpy array)

        rmin:
            minimal radius (inclusive)

        rmax:
            maximal radius (inclusive)

        kind:
            either ``"basic"`` (only VoM), or ``"normalized"`` (VoM/MoV);
            normalized version increases subpixel bias, but it works better at lower SNR

        highpass_size:
            size of the high-pass filter; ``None`` means no filter (effectively, infinite size)

        upsample: int
            integer image upsampling factor; `rmin` and `rmax` are adjusted
            automatically (i.e., they refer to the non-upsampled image); if
            ``upsample>1``, the resulting image size is multiplied by
            ``upsample``

        rweights:
            relative weights of different radial kernels; must be a 1D array of
            the length ``(rmax-rmin+1)//coarse_factor`` coarse_factor: the
            reduction factor for the number ring kernels; can be used to speed
            up calculations at the expense of precision coarse_mode: the
            reduction method; can be ``"add"`` (add ``coarse_factor`` rings in
            a row to get a thicker ring, which works better for smooth
            features), or ``"skip"`` (only keep on in ``coarse_factor`` rings,
            which works better for very fine features)

        coarse_factor:
            the reduction factor for the number ring kernels; can be used to
            speed up calculations at the expense of precision

        coarse_mode:
            the reduction method; can be ``"add"`` (add ``coarse_factor`` rings
            in a row to get a thicker ring, which works better for smooth
            features), or ``"skip"`` (only keep on in ``coarse_factor`` rings,
            which works better for very fine features)

        pad_mode:
            edge padding mode for convolutions; can be either one of modes
            accepted by ``np.pad`` (such as ``"constant"``, ``"reflect"``, or
            ``"edge"``), or ``"fast"``, which means faster no-padding (a
            combination of ``"wrap"`` and ``"constant"`` padding depending on
            the image size); ``"fast"`` mode works faster for smaller images
            and larger ``rmax``, but the border pixels (within ``rmax`` from
            the edge) are less reliable; note that the image mean is subtracted
            before processing, so ``pad_mode="constant"`` (default) is
            equivalent to padding with a constant value equal to the image mean

        Returns
        -------
            Returns transform source image

        """
        upsample = int(upsample)
        self._check_args(rmin, rmax, kind, coarse_mode, highpass_size, upsample)
        if highpass_size is not None:
            img = self.high_pass(img, highpass_size)
        if upsample > 1:
            img = img.repeat(upsample, axis=-2).repeat(
                upsample, axis=-1
            )  # nearest-neighbor upsampling on both axes
            if rweights is not None:
                rweights = np.asarray(rweights).repeat(
                    upsample
                )  # upsample radial weights as well
            rmin *= upsample  # increase minimal and maximal radii
            rmax *= upsample

        tmp = self.rvt_core(
            img,
            rmin,
            rmax,
            kind=kind,
            rweights=rweights,
            coarse_factor=coarse_factor,
            coarse_mode=coarse_mode,
            pad_mode=pad_mode,
        )
        return tmp

    def rvt_video(
        self,
        video,
        rmin,
        rmax,
        kind="basic",
        highpass_size=None,
        upsample=1,
        rweights=None,
        coarse_factor=1,
        coarse_mode="add",
        pad_mode="constant",
    ):
        """This is an RVT wrapper that allows you to get video in parallel.

        Parameters
        ----------
        video: NDArray
             The video is 3D-numpy (number of frames, width, height).

        rmin:
            minimal radius (inclusive)

        rmax:
            maximal radius (inclusive)

        kind:
            either ``"basic"`` (only VoM), or ``"normalized"`` (VoM/MoV);
            normalized version increases subpixel bias, but it works better at
            lower SNR

        highpass_size:
            size of the high-pass filter; ``None`` means no filter
            (effectively, infinite size)

        upsample: int
            integer image upsampling factor; `rmin` and `rmax` are adjusted
            automatically (i.e., they refer to the non-upsampled image); if
            ``upsample>1``, the resulting image size is multiplied by
            ``upsample``

        rweights:
            relative weights of different radial kernels; must be a 1D array of
            the length ``(rmax-rmin+1)//coarse_factor`` coarse_factor: the
            reduction factor for the number ring kernels; can be used to speed
            up calculations at the expense of precision coarse_mode: the
            reduction method; can be ``"add"`` (add ``coarse_factor`` rings in
            a row to get a thicker ring, which works better for smooth
            features), or ``"skip"`` (only keep on in ``coarse_factor`` rings,
            which works better for very fine features)

        coarse_factor:
            the reduction factor for the number ring kernels; can be used to
            speed up calculations at the expense of precision

        coarse_mode:
            the reduction method; can be ``"add"`` (add ``coarse_factor`` rings
            in a row to get a thicker ring, which works better for smooth
            features), or ``"skip"`` (only keep on in ``coarse_factor`` rings,
            which works better for very fine features)

        pad_mode:
            edge padding mode for convolutions; can be either one of modes
            accepted by ``np.pad`` (such as ``"constant"``, ``"reflect"``, or
            ``"edge"``), or ``"fast"``, which means faster no-padding (a
            combination of ``"wrap"`` and ``"constant"`` padding depending on
            the image size); ``"fast"`` mode works faster for smaller images
            and larger ``rmax``, but the border pixels (within ``rmax`` from
            the edge) are less reliable; note that the image mean is subtracted
            before processing, so ``pad_mode="constant"`` (default) is
            equivalent to padding with a constant value equal to the image mean

        Returns
        -------
            Returns transform source video.

        """
        self.video = video

        if self.cpu.parallel_active is True and self.inter_flag_parallel_active is True:
            print("\n---start RVT with Parallel---")

            result = Parallel(
                n_jobs=self.cpu.n_jobs, backend=self.cpu.backend, verbose=self.cpu.verbose
            )(
                delayed(self.rvt_kernel)(
                    f_,
                    rmin=rmin,
                    rmax=rmax,
                    kind=kind,
                    highpass_size=highpass_size,
                    upsample=upsample,
                    rweights=rweights,
                    coarse_factor=coarse_factor,
                    coarse_mode=coarse_mode,
                    pad_mode=pad_mode,
                )
                for f_ in tqdm(range(video.shape[0]))
            )
            arry_result = np.asarray(result)
            self.rvt_video = np.reshape(
                arry_result, (len(result), self.video.shape[1], self.video.shape[2])
            )
        else:
            print("\n---start RVT without Parallel---")

            result = [
                self.rvt_kernel(
                    i_,
                    rmin=rmin,
                    rmax=rmax,
                    kind=kind,
                    highpass_size=highpass_size,
                    upsample=upsample,
                    rweights=rweights,
                    coarse_factor=coarse_factor,
                    coarse_mode=coarse_mode,
                    pad_mode=pad_mode,
                )
                for i_ in tqdm(range(self.video.shape[0]))
            ]
            arry_result = np.asarray(result)
            self.rvt_video = np.reshape(
                arry_result, (len(result), self.video.shape[1], self.video.shape[2])
            )
        return self.rvt_video

    def rvt_kernel(
        self,
        frame_num,
        rmin,
        rmax,
        kind="basic",
        highpass_size=None,
        upsample=1,
        rweights=None,
        coarse_factor=1,
        coarse_mode="add",
        pad_mode="constant",
    ):
        """Perform Radial Variance Transform (RVT) of an image.

        Parameters
        ----------
        frame_num: int
            frame number

        rmin:
            minimal radius (inclusive)

        rmax:
            maximal radius (inclusive)

        kind:
            either ``"basic"`` (only VoM), or ``"normalized"`` (VoM/MoV);
            normalized version increases subpixel bias, but it works better at
            lower SNR

        highpass_size:
            size of the high-pass filter; ``None`` means no filter
            (effectively, infinite size)

        upsample: int
            integer image upsampling factor; `rmin` and `rmax` are adjusted
            automatically (i.e., they refer to the non-upsampled image); if
            ``upsample>1``, the resulting image size is multiplied by
            ``upsample``

        rweights:
            relative weights of different radial kernels; must be a 1D array of
            the length ``(rmax-rmin+1)//coarse_factor`` coarse_factor: the
            reduction factor for the number ring kernels; can be used to speed
            up calculations at the expense of precision coarse_mode: the
            reduction method; can be ``"add"`` (add ``coarse_factor`` rings in
            a row to get a thicker ring, which works better for smooth
            features), or ``"skip"`` (only keep on in ``coarse_factor`` rings,
            which works better for very fine features)

        coarse_factor:
            the reduction factor for the number ring kernels; can be used to
            speed up calculations at the expense of precision

        coarse_mode:
            the reduction method; can be ``"add"`` (add ``coarse_factor`` rings
            in a row to get a thicker ring, which works better for smooth
            features), or ``"skip"`` (only keep on in ``coarse_factor`` rings,
            which works better for very fine features)

        pad_mode:
            edge padding mode for convolutions; can be either one of modes
            accepted by ``np.pad`` (such as ``"constant"``, ``"reflect"``, or
            ``"edge"``), or ``"fast"``, which means faster no-padding (a
            combination of ``"wrap"`` and ``"constant"`` padding depending on
            the image size); ``"fast"`` mode works faster for smaller images
            and larger ``rmax``, but the border pixels (within ``rmax`` from
            the edge) are less reliable; note that the image mean is subtracted
            before processing, so ``pad_mode="constant"`` (default) is
            equivalent to padding with a constant value equal to the image mean

        Returns
        -------
            Returns transform source image

        """
        img = self.video[frame_num, :, :]
        upsample = int(upsample)
        self._check_args(rmin, rmax, kind, coarse_mode, highpass_size, upsample)
        if highpass_size is not None:
            img = self.high_pass(img, highpass_size)
        if upsample > 1:
            img = img.repeat(upsample, axis=-2).repeat(
                upsample, axis=-1
            )  # nearest-neighbor upsampling on both axes
            if rweights is not None:
                rweights = np.asarray(rweights).repeat(
                    upsample
                )  # upsample radial weights as well
            rmin *= upsample  # increase minimal and maximal radii
            rmax *= upsample
        return self.rvt_core(
            img,
            rmin,
            rmax,
            kind=kind,
            rweights=rweights,
            coarse_factor=coarse_factor,
            coarse_mode=coarse_mode,
            pad_mode=pad_mode,
        )


class FastRadialSymmetryTransform:
    def __init__(self):
        """Implementation of fast radial symmetry transform in pure Python using OpenCV and numpy.

        References
        ----------
        The fast radial symmetry transform code has been adopted from the GitHub repository mentioned below.

            [1] Loy, G., & Zelinsky, A. (2002). A fast radial symmetry transform
            for detecting points of interest. Computer Vision, ECCV 2002.

            [2] https://github.com/Xonxt/frst

        """
        pass

    def gradx(self, img):
        rows, cols = img.shape
        return np.hstack(
            (np.zeros((rows, 1)), (img[:, 2:] - img[:, :-2]) / 2.0, np.zeros((rows, 1)))
        )

    def grady(self, img):
        # img = img.astype('int')
        rows, cols = img.shape
        # Use vstack to add back the rows that were dropped as zeros
        return np.vstack(
            (np.zeros((1, cols)), (img[2:, :] - img[:-2, :]) / 2.0, np.zeros((1, cols)))
        )

    def _frst(self, img, radii, alpha, beta, stdFactor, mode="BOTH"):
        """Performs fast radial symmetry transform

        Parameters
        ----------
        img: NDArray
           Input_video image, grayscale.

        radii: int
           Integer value for radius size in pixels (n in the original paper);
           also used to size gaussian kernel

        alpha: float
           Strictness of symmetry transform (higher=more strict; 2 is good
           place to start)

        beta: float
            Gradient threshold_min parameter, float in [0,1]

        stdFactor: float
           Standard deviation factor for gaussian kernel

        mode: str
           BRIGHT, DARK, or BOTH

        """
        mode = mode.upper()
        assert mode in ["BRIGHT", "DARK", "BOTH"]
        dark = mode == "DARK" or mode == "BOTH"
        bright = mode == "BRIGHT" or mode == "BOTH"

        workingDims = tuple((e + 2 * radii) for e in img.shape)

        output = np.zeros(img.shape, np.float64)
        O_n = np.zeros(workingDims, np.float64)
        M_n = np.zeros(workingDims, np.float64)

        # Calculate gradients
        gx = self.gradx(img)
        gy = self.grady(img)

        # Find gradient vector magnitude
        gnorms = np.sqrt(np.add(np.multiply(gx, gx), np.multiply(gy, gy)))

        # Use beta to set threshold_min - speeds up transform significantly
        gthresh = np.amax(gnorms) * beta

        # Find x/y distance to affected pixels
        gpx = (
            np.multiply(np.divide(gx, gnorms, out=np.zeros(gx.shape), where=gnorms != 0), radii)
            .round()
            .astype(int)
        )
        gpy = (
            np.multiply(np.divide(gy, gnorms, out=np.zeros(gy.shape), where=gnorms != 0), radii)
            .round()
            .astype(int)
        )

        # Iterate over all pixels (w/ gradient above threshold_min)
        for coords, gnorm in np.ndenumerate(gnorms):
            if gnorm > gthresh:
                i, j = coords
                # Positively affected pixel
                if bright:
                    ppve = (i + gpx[i, j], j + gpy[i, j])
                    O_n[ppve] += 1
                    M_n[ppve] += gnorm
                # Negatively affected pixel
                if dark:
                    pnve = (i - gpx[i, j], j - gpy[i, j])
                    O_n[pnve] -= 1
                    M_n[pnve] -= gnorm

        # Abs and normalize O matrix
        O_n = np.abs(O_n)
        O_n = O_n / float(np.amax(O_n))

        # Normalize M matrix
        M_max = float(np.amax(np.abs(M_n)))
        M_n = M_n / M_max

        # Elementwise multiplication
        F_n = np.multiply(np.power(O_n, alpha), M_n)

        # Gaussian blur
        kSize = int(np.ceil(radii / 2))
        kSize = kSize + 1 if kSize % 2 == 0 else kSize

        S = cv2.GaussianBlur(F_n, (kSize, kSize), int(radii * stdFactor))
        S = np.fliplr(S)

        return S[radii:-radii, radii:-radii]


class GuidedFilter:
    def __init__(self, I, radius, eps):
        """
        This is a class which builds guided filter
        according to the channel number of guided Input.
        The guided input could be gray image, color image,
        or multi-dimensional feature map.

        Parameters
        ----------
        I: NDArray
            Guided image or guided feature map

        radius: int
            Radius of filter

        eps: float
            Value controlling sharpness

        References
        ----------
          [1] K.He, J.Sun, and X.Tang. Guided Image Filtering. TPAMI'12.
        """
        if len(I.shape) == 2:
            self._Filter = GrayGuidedFilter(I, radius, eps)
        else:
            self._Filter = MultiDimGuidedFilter(I, radius, eps)

    def filter(self, p):
        """
        Parameters
        ----------
        p: NDArray
            Filtering input which is 2D or 3D with format
            HW or HWC

        Returns
        -------
        ret: NDArray
            Filtering output whose shape is same with input
        """
        p = np.float32(p)
        if len(p.shape) == 2:
            return self._Filter.filter(p)
        elif len(p.shape) == 3:
            channels = p.shape[2]
            ret = np.zeros_like(p, dtype=np.float32)
            for c in range(channels):
                ret[:, :, c] = self._Filter.filter(p[:, :, c])
            return ret


class GrayGuidedFilter:
    def __init__(self, I, radius, eps):
        """
        Specific guided filter for gray guided image.

        Parameters
        ----------
        I: NDArray
            2D guided image

        radius: int
            Radius of filter

        eps: float
            Value controlling sharpness

        References
        ----------
        The guided filter for gray guided image code has been adopted from the GitHub repository mentioned below.

            [1] https://github.com/lisabug/guided-filter
        """
        self.I = np.float32(I)
        self.radius = radius
        self.eps = eps

    def filter(self, p):
        """
        Parameters
        ----------
        p: NDArray
            Filtering input of 2D

        Returns
        -------
        q: NDArray
            Filtering output of 2D
        """
        # step 1
        meanI = cv2.boxFilter(src=self.I, ddepth=-1, ksize=self.radius)
        meanp = cv2.boxFilter(src=p, ddepth=-1, ksize=self.radius)
        corrI = cv2.boxFilter(src=self.I * self.I, ddepth=-1, ksize=self.radius)
        corrIp = cv2.boxFilter(src=self.I * p, ddepth=-1, ksize=self.radius)
        # step 2
        varI = corrI - meanI * meanI
        covIp = corrIp - meanI * meanp
        # step 3
        a = covIp / (varI + self.eps)
        b = meanp - a * meanI
        # step 4
        meana = cv2.boxFilter(src=a, ddepth=-1, ksize=self.radius)
        meanb = cv2.boxFilter(src=b, ddepth=-1, ksize=self.radius)
        # step 5
        q = meana * self.I + meanb

        return q


class MultiDimGuidedFilter:
    def __init__(self, I, radius, eps):
        """
        Specific guided filter for color guided image or multi-dimensional feature map.

        Parameters
        ----------
        I: NDArray
            Image.

        radius: int
            Radius of filter

        eps: float
            Value controlling sharpness

        References
        ----------

        The guided filter for color guided image code has been adopted from the GitHub repository mentioned below.

            [1] https://github.com/lisabug/guided-filter
        """
        self.I = np.float32(I)
        self.radius = radius
        self.eps = eps

        self.rows = self.I.shape[0]
        self.cols = self.I.shape[1]
        self.chs = self.I.shape[2]

    def filter(self, p):
        """
        Parameters
        ----------
        p: NDArray
            Filtering input of 2D

        Returns
        -------
        q: NDArray
            Filtering output of 2D
        """
        p_ = np.expand_dims(p, axis=2)

        meanI = cv2.boxFilter(src=self.I, ksize=self.radius)  # (H, W, C)
        meanp = cv2.boxFilter(src=p_, ksize=self.radius)  # (H, W, 1)
        I_ = self.I.reshape((self.rows * self.cols, self.chs, 1))  # (HW, C, 1)
        meanI_ = meanI.reshape((self.rows * self.cols, self.chs, 1))  # (HW, C, 1)

        corrI_ = np.matmul(I_, I_.transpose(0, 2, 1))  # (HW, C, C)
        corrI_ = corrI_.reshape((self.rows, self.cols, self.chs * self.chs))  # (H, W, CC)
        corrI_ = cv2.boxFilter(src=corrI_, ksize=self.radius)
        corrI = corrI_.reshape((self.rows * self.cols, self.chs, self.chs))  # (HW, C, C)

        U = np.expand_dims(np.eye(self.chs, dtype=np.float32), axis=0)
        # U = np.tile(U, (self.rows*self.cols, 1, 1)) # (HW, C, C)

        left = np.linalg.inv(corrI + self.eps * U)  # (HW, C, C)

        corrIp = cv2.boxFilter(src=self.I * p_, ksize=self.radius)  # (H, W, C)
        covIp = corrIp - meanI * meanp  # (H, W, C)
        right = covIp.reshape((self.rows * self.cols, self.chs, 1))  # (HW, C, 1)

        a = np.matmul(left, right)  # (HW, C, 1)
        axmeanI = np.matmul(a.transpose((0, 2, 1)), meanI_)  # (HW, 1, 1)
        axmeanI = axmeanI.reshape((self.rows, self.cols, 1))
        b = meanp - axmeanI  # (H, W, 1)
        a = a.reshape((self.rows, self.cols, self.chs))

        meana = cv2.boxFilter(src=a, ksize=self.radius)
        meanb = cv2.boxFilter(src=b, ksize=self.radius)

        meana = meana.reshape((self.rows * self.cols, 1, self.chs))
        meanb = meanb.reshape((self.rows * self.cols, 1, 1))
        I_ = self.I.reshape((self.rows * self.cols, self.chs, 1))

        q = np.matmul(meana, I_) + meanb
        q = q.reshape((self.rows, self.cols))

        return q



"""
__author__ = "Houman Mirzaalian D., xxxx"
"""

from __future__ import print_function

import numpy as np
from matplotlib import animation
from matplotlib import pyplot as plt
from matplotlib.patches import Circle, Rectangle
from mpl_toolkits.axes_grid1 import make_axes_locatable
from PySide6.QtCore import QObject, QRunnable, Signal, Slot
from scipy.ndimage import median_filter

from piscat.InputOutput import read_status_line
from piscat.Trajectory.data_handling import protein_trajectories_list2dic
from piscat.Visualization.print_colors import PrintColors


class WorkerSignals(QObject):
    finished = Signal()
    error = Signal(tuple)
    result = Signal(object)
    progress = Signal(int)


class HTML_Display:
    def __init__(
        self,
        video,
        color="gray",
        time_delay=0.5,
        median_filter_flag=False,
        imgSizex=5,
        imgSizey=5,
    ):
        """This class display the video for HTML.

        Parameters
        ----------
        video:(NDArray)
           Input video

        median_filter_flag: bool
           In case it defines as True, a median filter is applied with size 3
           to remove hot pixel effect.

        color: str
           It defines the colormap for visualization.

        time_delay: float
           Delay between frames in milliseconds.

        imgSizex: int
          Image length size.

        imgSizey: int
          Image width size.

        """
        self.video = video
        self.median_filter_flag = median_filter_flag

        self.fig = plt.figure()
        self.fig.set_size_inches(imgSizex, imgSizey, True)
        self.ax = self.fig.add_subplot(111)
        self.pressed_key = {}
        plt.axis("off")

        self.div = make_axes_locatable(self.ax)
        self.cax = self.div.append_axes("right", "5%", "5%")

        self.cv0 = self.video[0, :, :]
        self.im = self.ax.imshow(self.cv0, origin="lower", cmap=color)
        self.cb = self.fig.colorbar(self.im, cax=self.cax)

        self.tx = self.ax.set_title("Frame 0")

        self.ani = animation.FuncAnimation(
            self.fig,
            self.animate,
            frames=self.video.shape[0],
            blit=False,
            interval=time_delay,
            repeat=False,
            cache_frame_data=False,
        )

    def animate(self, i_):
        arr = self.video[i_, :, :]
        vmax = np.max(arr)
        vmin = np.min(arr)
        if self.median_filter_flag:
            frame_v = median_filter(arr, 3)
        else:
            frame_v = arr
        self.im.set_data(frame_v)
        self.im.set_clim(vmin, vmax)
        self.tx.set_text("Frame {0}".format(i_))


class HTML_Display_StatusLine:
    def __init__(
        self,
        video,
        median_filter_flag=False,
        color="gray",
        imgSizex=5,
        imgSizey=5,
        time_delay=0.5,
    ):
        """This class displays the video in the HTML while highlight status line

        Parameters
        ----------
        video: NDArray
           Input video.

        median_filter_flag: bool
           In case it defines as True, a median filter is applied with size 3
           to remove hot pixel effect.

        color: str
           It defines the colormap for visualization.

        imgSizex: int
           Image length size.

        imgSizey: int
           Image width size.

        time_delay: float
           Delay between frames in milliseconds.

        """
        self.color = color
        self.video = video
        self.imgSizex = imgSizex
        self.imgSizey = imgSizey
        self.median_filter_flag = median_filter_flag

        status_ = read_status_line.StatusLine(video)
        _, status_information = status_.find_status_line()
        self.statusLine_position = status_information["status_line_position"]

        self.fig = plt.figure()
        self.fig.set_size_inches(imgSizex, imgSizey, True)
        self.ax = self.fig.add_subplot(111)
        self.pressed_key = {}
        plt.axis("off")

        self.div = make_axes_locatable(self.ax)
        self.cax = self.div.append_axes("right", "5%", "5%")

        self.cv0 = self.video[0, :, :]
        self.im = self.ax.imshow(self.cv0, origin="lower", cmap=self.color)
        self.cb = self.fig.colorbar(self.im, cax=self.cax)
        self.tx = self.ax.set_title("Frame 0")

        if self.statusLine_position == "column":
            rect = Rectangle(
                (0, self.video.shape[2]),
                self.video.shape[2],
                1,
                linewidth=10,
                edgecolor="r",
                facecolor="none",
            )
            self.ax.add_patch(rect)
        elif self.statusLine_position == "row":
            rect = Rectangle(
                (self.video.shape[1], 0),
                1,
                self.video.shape[1],
                linewidth=10,
                edgecolor="r",
                facecolor="none",
            )
            self.ax.add_patch(rect)

        self.ani = animation.FuncAnimation(
            self.fig,
            self.animate,
            frames=self.video.shape[0],
            blit=False,
            interval=time_delay,
            repeat=False,
            cache_frame_data=False,
        )

    def animate(self, i_):
        [p.remove() for p in reversed(self.ax.patches)]
        arr = self.video[i_, :, :]
        vmax = np.max(arr)
        vmin = np.min(arr)
        if self.median_filter_flag:
            frame_v = median_filter(arr, 3)
        else:
            frame_v = arr
        self.im.set_data(frame_v)
        self.im.set_clim(vmin, vmax)
        self.tx.set_text("Frame {0}".format(i_))

        if self.statusLine_position == "column":
            rect = Rectangle(
                (0, self.video.shape[2]),
                self.video.shape[2],
                1,
                linewidth=10,
                edgecolor="r",
                facecolor="none",
            )
            self.ax.add_patch(rect)

        elif self.statusLine_position == "row":
            rect = Rectangle(
                (self.video.shape[1], 0),
                1,
                self.video.shape[1],
                linewidth=10,
                edgecolor="r",
                facecolor="none",
            )
            self.ax.add_patch(rect)


class HTML_PSFs_subplotLocalizationDisplay(QRunnable, PrintColors):
    def __init__(
        self,
        list_videos,
        list_df_PSFs,
        list_titles,
        numRows,
        numColumns,
        color="gray",
        median_filter_flag=False,
        imgSizex=5,
        imgSizey=5,
        time_delay=0.5,
        save_path=None,
        fps=10,
    ):
        """
        This class displays the videos in the HTML while highlight PSFs.

        Parameters
        ----------
        list_videos: list of NDArray
            List of videos.

        list_df_PSFs: list panda data_frame
            List Data Frames that contains the location of PSFs for each video.

        numRows: int
            It defines number of rows in sub-display.

        numColumns: int
            It defines number of columns in sub-display.

        list_titles: list str
            List of titles for each sub plot.

        median_filter_flag: bool
            In case it defines as True, a median filter is applied with size 3
            to remove hot pixel effect.

        color: str
            It defines the colormap for visualization.

        imgSizex: int
            Image length size.

        imgSizey: int
            Image width size.

        time_delay: float
                   Delay between frames in milliseconds.
        """
        super(HTML_PSFs_subplotLocalizationDisplay, self).__init__()
        PrintColors.__init__(self)

        self.list_video = list_videos
        self.list_df_PSFs = list_df_PSFs
        self.save_path = save_path
        self.fps = fps
        self.signals = WorkerSignals()
        self.median_filter_flag = median_filter_flag
        self.memory = 0
        self.numRows = numRows
        self.numColumns = numColumns
        self.numVideos = len(list_videos)
        if list_titles is None:
            self.list_titles = [None for _ in range(self.numVideos)]
        else:
            self.list_titles = list_titles

        self.fig = plt.figure(figsize=(imgSizex, imgSizey))
        self.imgGrid_list = []
        for i_ in range(1, self.numRows * self.numColumns + 1):
            img_grid_ = self.fig.add_subplot(self.numRows, self.numColumns, i_)
            img_grid_.axis("off")
            self.imgGrid_list.append(img_grid_)

        self.fig.tight_layout()
        self.fig.subplots_adjust(wspace=0.18)

        self.div_0 = []
        self.cax_0 = []
        self.im_0 = []
        self.cb_0 = []
        self.tx_0 = []

        for ax_, vid_, df_PSFs, tit_ in zip(
            self.imgGrid_list, self.list_video, self.list_df_PSFs, self.list_titles
        ):
            div_ = make_axes_locatable(ax_)
            self.div_0.append(div_)

            cax_ = div_.append_axes("right", "5%", "5%")
            self.cax_0.append(cax_)

            im_ = ax_.imshow(vid_[0, :, :], origin="lower", cmap=color)
            self.im_0.append(im_)
            self.cb_0.append(self.fig.colorbar(im_, cax=cax_))

            if tit_ is not None:
                self.tx_0.append(ax_.set_title(tit_ + ", Frame 0"))
            else:
                self.tx_0.append(ax_.set_title("Frame 0"))

            particle = df_PSFs.loc[df_PSFs["frame"] == 0]
            particle_X = particle["x"].tolist()
            particle_Y = particle["y"].tolist()
            particle_sigma = particle["sigma"].tolist()

            for j_ in range(len(particle_X)):
                y = int(particle_Y[j_])
                x = int(particle_X[j_])
                sigma = particle_sigma[j_]
                ax_.add_patch(
                    Circle(
                        (x, y),
                        radius=np.sqrt(2) * sigma,
                        edgecolor="r",
                        facecolor="none",
                        linewidth=2,
                    )
                )

        self.ani = animation.FuncAnimation(
            self.fig,
            self.animate,
            frames=self.list_video[0].shape[0],
            blit=False,
            interval=time_delay,
            repeat=False,
            cache_frame_data=False,
        )

    def animate(self, i_):
        if i_ < self.list_video[0].shape[0]:
            for idx_, (ax_, vid_, df_PSFs, tit_) in enumerate(
                zip(self.imgGrid_list, self.list_video, self.list_df_PSFs, self.list_titles)
            ):
                [p.remove() for p in reversed(ax_.patches)]

                arr = vid_[i_, :, :]
                vmax = np.max(arr)
                vmin = np.min(arr)
                if self.median_filter_flag:
                    frame_v = median_filter(arr, 3)
                else:
                    frame_v = arr

                self.im_0[idx_].set_data(frame_v)
                self.im_0[idx_].set_clim(vmin, vmax)

                if tit_ is not None:
                    self.tx_0[idx_].set_text(tit_ + ", Frame {}".format(i_))
                else:
                    self.tx_0[idx_].set_text("Frame {}".format(i_))

                particle = df_PSFs.loc[df_PSFs["frame"] == i_]
                particle_X = particle["x"].tolist()
                particle_Y = particle["y"].tolist()
                particle_sigma = particle["sigma"].tolist()
                for j_ in range(len(particle_X)):
                    y = int(particle_Y[j_])
                    x = int(particle_X[j_])
                    sigma = particle_sigma[j_]
                    ax_.add_patch(
                        Circle(
                            (x, y),
                            radius=np.sqrt(2) * sigma,
                            edgecolor="r",
                            facecolor="none",
                            linewidth=2,
                        )
                    )

    @Slot()
    def run(self, *args, **kwargs):
        if self.save_path is not None:
            dpi = 300
            writer = animation.writers["ffmpeg"](
                fps=self.fps, bitrate=-1, extra_args=["-vcodec", "libx264"]
            )
            self.ani.save(self.save_path, writer=writer, dpi=dpi)

            self.signals.finished.emit()
        else:
            print(f"{self.WARNING}\nError occurred in MP4!{self.ENDC}")
            self.signals.finished.emit()


class HTML_subplotDisplay:
    def __init__(
        self,
        list_videos,
        list_titles,
        numRows,
        numColumns,
        color="gray",
        median_filter_flag=False,
        imgSizex=5,
        imgSizey=5,
        time_delay=0.5,
    ):
        """
        This class displays the videos in the HTML.

        Parameters
        ----------
        list_videos: list of NDArray
            List of videos.

        numRows: int
            It defines number of rows in sub-display.

        numColumns: int
            It defines number of columns in sub-display.

        list_titles: list str
            List of titles for each sub plot.

        median_filter_flag: bool
            In case it defines as True, a median filter is applied with size 3
            to remove hot pixel effect.

        color: str
            It defines the colormap for visualization.

        imgSizex: int
            Image length size.

        imgSizey: int
            Image width size.

        time_delay: float
            Delay between frames in milliseconds.
        """

        self.list_video = list_videos
        self.median_filter_flag = median_filter_flag
        self.memory = 0
        self.numRows = numRows
        self.numColumns = numColumns
        self.numVideos = len(list_videos)
        if list_titles is None:
            self.list_titles = [None for _ in range(self.numVideos)]
        else:
            self.list_titles = list_titles

        self.fig = plt.figure(figsize=(imgSizex, imgSizey))
        self.imgGrid_list = []
        for i_ in range(1, self.numRows * self.numColumns + 1):
            img_grid_ = self.fig.add_subplot(self.numRows, self.numColumns, i_)
            img_grid_.axis("off")
            self.imgGrid_list.append(img_grid_)

        self.fig.tight_layout()
        self.fig.subplots_adjust(wspace=0.18)

        self.div_0 = []
        self.cax_0 = []
        self.im_0 = []
        self.cb_0 = []
        self.tx_0 = []

        for ax_, vid_, tit_ in zip(self.imgGrid_list, self.list_video, self.list_titles):
            div_ = make_axes_locatable(ax_)
            self.div_0.append(div_)

            cax_ = div_.append_axes("right", "5%", "5%")
            self.cax_0.append(cax_)

            im_ = ax_.imshow(vid_[0, :, :], origin="lower", cmap=color)
            self.im_0.append(im_)
            self.cb_0.append(self.fig.colorbar(im_, cax=cax_))

            if tit_ is not None:
                self.tx_0.append(ax_.set_title(tit_ + ", Frame 0"))
            else:
                self.tx_0.append(ax_.set_title("Frame 0"))

        self.ani = animation.FuncAnimation(
            self.fig,
            self.animate,
            frames=self.list_video[0].shape[0],
            blit=False,
            interval=time_delay,
            repeat=False,
            cache_frame_data=False,
        )

    def animate(self, i_):
        if i_ < self.list_video[0].shape[0]:
            for idx_, (_, vid_, tit_) in enumerate(
                zip(self.imgGrid_list, self.list_video, self.list_titles)
            ):
                arr = vid_[i_, :, :]
                vmax = np.max(arr)
                vmin = np.min(arr)
                if self.median_filter_flag:
                    frame_v = median_filter(arr, 3)
                else:
                    frame_v = arr

                self.im_0[idx_].set_data(frame_v)
                self.im_0[idx_].set_clim(vmin, vmax)

                if tit_ is not None:
                    self.tx_0[idx_].set_text(tit_ + ", Frame {}".format(i_))
                else:
                    self.tx_0[idx_].set_text("Frame {}".format(i_))


class HTMLSelectedPSFs_localizationDisplay:
    def __init__(
        self,
        video,
        particles,
        particles_num="#0",
        frame_extend=0,
        color="gray",
        median_filter_flag=False,
        imgSizex=5,
        imgSizey=5,
        time_delay=0.5,
    ):
        """This class displays video in the HTML while highlighting selected
        PSF.

        Parameters
        ----------
        video: NDArray
           Input video.

        particles: dic
             Dictionary similar to the following structures.:

            | {"#0": {'intensity_horizontal': ..., 'intensity_vertical': ...,
               ..., 'particle_ID': ...}, "#1": {}, ...}

        particles_num: str
            Choose the corresponding key in the particles dictionary.

        frame_extend: int
            Display particle for ``frame_extend`` before and after segmented
            ones. In case there are not enough frames before/after, it shows
            only for the number of existing frames.

        flag_fit2D: bool
            It activate 2D-Gaussian fit to extract fitting information of
            selected PSF.

        median_filter_flag: bool
           In case it defines as True, a median filter is applied with size 3
           to remove hot pixel effect.

        color: str
           It defines the colormap for visualization.

        imgSizex: int
           Image length size.

        imgSizey: int
           Image width size.

        time_delay: float
            Delay between frames in milliseconds.

        """

        self.video = video
        self.particles = particles
        self.median_filter_flag = median_filter_flag
        self.memory = 0

        if type(particles) is list:
            self.particles = protein_trajectories_list2dic(particles)

        if type(self.particles) is dict:
            for key, item in self.particles.items():
                if key == particles_num:
                    intensity_horizontal = item["intensity_horizontal"]
                    intensity_vertical = item["intensity_vertical"]
                    center_int = item["center_int"]
                    center_int_flow = item["center_int_flow"]
                    self.frame_number_ = item["frame_number"]
                    self.sigma = item["sigma"]
                    self.x_center = item["x_center"]
                    self.y_center = item["y_center"]
                    self.particle_ID = item["particle_ID"]

        video_frameNum = video.shape[0]
        max_particle_frame = np.max(self.frame_number_)
        min_particle_frame = np.min(self.frame_number_)

        max_extend_particle_frame = np.min([max_particle_frame + frame_extend, video_frameNum])
        min_extend_particle_frame = np.max([min_particle_frame - frame_extend, 0])

        self.frame_number = list(range(min_extend_particle_frame, max_extend_particle_frame, 1))

        self.fig = plt.figure(figsize=(imgSizex, imgSizey))
        self.fig.subplots_adjust(left=0, bottom=0, right=1, top=1, wspace=None, hspace=None)
        self.imgGrid_list = []

        img_grid_ = self.fig.add_subplot(1, 1, 1)
        img_grid_.axis("off")
        self.imgGrid_list = img_grid_

        self.fig.tight_layout()
        # self.fig.subplots_adjust(wspace=0.18)

        self.div_0 = []
        self.cax_0 = []
        self.im_0 = []
        self.cb_0 = []
        self.tx_0 = []

        div_ = make_axes_locatable(self.imgGrid_list)
        self.div_0.append(div_)

        cax_ = div_.append_axes("right", "5%", "5%")
        self.cax_0.append(cax_)

        selected_frame = self.frame_number[0]
        im_ = self.imgGrid_list.imshow(video[selected_frame, :, :], origin="lower", cmap=color)
        self.im_0 = im_
        self.cb_0 = self.fig.colorbar(im_, cax=cax_)

        self.tx_0 = self.imgGrid_list.set_title("Frame 0")

        if 0 <= np.max(self.frame_number_) and 0 >= np.min(self.frame_number_):
            idx_ = np.where(self.frame_number_ == 0)

            particle_X = self.x_center[idx_]
            particle_Y = self.y_center[idx_]
            particle_sigma = self.sigma[idx_]
            particle_labels = self.particle_ID[idx_]

            y = int(particle_Y)
            x = int(particle_X)
            sigma = particle_sigma
            self.imgGrid_list.add_patch(
                Circle(
                    (x, y),
                    radius=np.sqrt(2) * sigma,
                    edgecolor="r",
                    facecolor="none",
                    linewidth=2,
                )
            )

        self.ani = animation.FuncAnimation(
            self.fig,
            self.animate,
            frames=self.frame_number,
            blit=False,
            interval=time_delay,
            repeat=False,
            cache_frame_data=False,
        )

    def animate(self, i_):
        if i_ <= np.max(self.frame_number) and i_ >= np.min(self.frame_number):
            [p.remove() for p in reversed(self.imgGrid_list.patches)]

            arr = self.video[i_, :, :]
            vmax = np.max(arr)
            vmin = np.min(arr)
            if self.median_filter_flag:
                frame_v = median_filter(arr, 3)
            else:
                frame_v = arr

            self.im_0.set_data(frame_v)
            self.im_0.set_clim(vmin, vmax)
            self.tx_0.set_text("Frame {}".format(i_))

            if i_ <= np.max(self.frame_number_) and i_ >= np.min(self.frame_number_):
                idx_ = np.where(self.frame_number_ == i_)

                particle_X = self.x_center[idx_]
                particle_Y = self.y_center[idx_]
                particle_sigma = self.sigma[idx_]
                particle_labels = self.particle_ID[idx_]

                y = int(particle_Y)
                x = int(particle_X)
                sigma = particle_sigma
                self.imgGrid_list.add_patch(
                    Circle(
                        (x, y),
                        radius=np.sqrt(2) * sigma,
                        edgecolor="r",
                        facecolor="none",
                        linewidth=2,
                    )
                )



import cmath
import math

import numpy as np
import scipy.special

from piscat.iPSF_model.L_theta import L_theta


class VectorialDiffractionIntegral:
    def __init__(self, p, Xp=(0, 0, 0), z_=0, nz=1, nx=513):
        """A vectorial diffraction formulation to model the scattered light
        from a nanoparticle.

        References
        ----------
        "Point spread function in interferometric scattering microscopy
        (iSCAT). Part I: aberrations in defocusing and axial localization.",
        Reza Gholami Mahmoodabadi, Richard W. Taylor, Martin Kaller, Susann
        Spindler, Mahdi Mazaheri, Kiarash Kasaian, and Vahid Sandoghdar Optics
        Express Vol. 28, Issue 18, pp. 25969-25988 (2020)
        https://doi.org/10.1364/OE.401374

        Parameters
        ----------
        p : Imaging setup parameters which is an object from the class imagingSetupParameters.
        Xp : The 3D position of the scatterer in the sample, in unit of meters.
        z_ : The position of the focus. This could be an array as well.
        nz : The number of focus positions to calculate.
        nx : The lateral extent of the image given in pixels.

        """
        self.p = p
        self.xp_ = Xp[0]
        self.yp_ = Xp[1]
        self.zp_ = Xp[2]
        self.z_ = z_
        self.nz_ = nz
        self.nx = nx
        xystep_ = p.pixel_size_physical / p.M
        self.xystep_ = xystep_
        xymax_ = (nx - 1) / 2
        xymax_ = int(xymax_)
        self.xymax_ = xymax_

        self.xp_ *= 1 / xystep_
        self.yp_ *= 1 / xystep_

        rn = 1 + int(math.sqrt(self.xp_**2 + self.yp_**2))
        rn = int(rn)

        rmax_ = math.ceil(math.sqrt(2.0) * xymax_) + rn + 1
        self.rmax_ = rmax_
        npx_ = (2 * xymax_ + 1) ** 2
        npx_ = int(npx_)

        N = nx * nx * nz
        self.I0_int_re_op_1D = np.zeros((N))
        self.I1_int_re_op_1D = np.zeros((N))
        self.I2_int_re_op_1D = np.zeros((N))
        self.I0_int_im_op_1D = np.zeros((N))
        self.I1_int_im_op_1D = np.zeros((N))
        self.I2_int_im_op_1D = np.zeros((N))

        self.I0_int_re_op_3D = np.zeros((nz, nx, nx))
        self.I1_int_re_op_3D = np.zeros((nz, nx, nx))
        self.I2_int_re_op_3D = np.zeros((nz, nx, nx))
        self.I0_int_im_op_3D = np.zeros((nz, nx, nx))
        self.I1_int_im_op_3D = np.zeros((nz, nx, nx))
        self.I2_int_im_op_3D = np.zeros((nz, nx, nx))

        self.I0_int_re = np.zeros((nz, rmax_))
        self.I1_int_re = np.zeros((nz, rmax_))
        self.I2_int_re = np.zeros((nz, rmax_))
        self.I0_int_im = np.zeros((nz, rmax_))
        self.I1_int_im = np.zeros((nz, rmax_))
        self.I2_int_im = np.zeros((nz, rmax_))

    def calculate(self):
        p = self.p
        phC = 0

        ci = self.zp_ * (1 - p.ni / p.ns) + p.ni * (p.tg0 / p.ng0 + p.ti0 / p.ni0 - p.tg / p.ng)
        ud = 3
        L_th = np.zeros((2), dtype=complex)

        for k in range(0, self.nz_):  # defocusing loop
            L_th = L_theta(L_th, p.alpha, p, ci, self.z_[k], self.zp_)
            w_exp = np.abs(L_th[1])  # missing p.k0, multiply it into it later
            cst = 0.975
            while cst >= 0.9:
                L_th = L_theta(L_th, cst * p.alpha, p, ci, self.z_[k], self.zp_)
                if np.abs(L_th[1]) > w_exp:
                    w_exp = np.abs(L_th[1])
                cst -= 0.025

            w_exp *= p.k0

            for ri in range(0, self.rmax_):
                r = self.xystep_ * float(ri)
                constJ = p.k0 * r * p.ni

                if w_exp > constJ:
                    nSamples = 4 * int(1 + p.alpha * w_exp / math.pi)
                else:
                    nSamples = 4 * int(1 + p.alpha * constJ / math.pi)

                if nSamples < 20:
                    nSamples = 20

                step = p.alpha / float(nSamples)
                iconst = step / ud

                # Odd and evens of Simpson's rule
                sum_I0_even, sum_I1_even, sum_I2_even = self.simpsonsRuleVec(
                    ci, constJ, k, nSamples, p, phC, step, 0
                )
                sum_I0_odd, sum_I1_odd, sum_I2_odd = self.simpsonsRuleVec(
                    ci, constJ, k, nSamples, p, phC, step, 1
                )

                sum_I0_edge, sum_I1_edge, sum_I2_edge = self.theta_equal_to_alpha(
                    ci, constJ, k, p, phC
                )

                sum_I0 = sum_I0_even + sum_I0_odd + sum_I0_edge
                sum_I1 = sum_I1_even + sum_I1_odd + sum_I1_edge
                sum_I2 = sum_I2_even + sum_I2_odd + sum_I2_edge

                # Integrals complexValue
                self.I0_int_re[k, ri] = (sum_I0 * iconst).real
                self.I1_int_re[k, ri] = (sum_I1 * iconst).real
                self.I2_int_re[k, ri] = (sum_I2 * iconst).real
                self.I0_int_im[k, ri] = (sum_I0 * iconst).imag
                self.I1_int_im[k, ri] = (sum_I1 * iconst).imag
                self.I2_int_im[k, ri] = (sum_I2 * iconst).imag

            # Interpolate and fill up output arrays
            self.interpolation_xy_vec(k)

        self.I0_int_re_op_3D = np.reshape(self.I0_int_re_op_1D, (self.nz_, self.nx, self.nx))
        self.I1_int_re_op_3D = np.reshape(self.I1_int_re_op_1D, (self.nz_, self.nx, self.nx))
        self.I2_int_re_op_3D = np.reshape(self.I2_int_re_op_1D, (self.nz_, self.nx, self.nx))
        self.I0_int_im_op_3D = np.reshape(self.I0_int_im_op_1D, (self.nz_, self.nx, self.nx))
        self.I1_int_im_op_3D = np.reshape(self.I1_int_im_op_1D, (self.nz_, self.nx, self.nx))
        self.I2_int_im_op_3D = np.reshape(self.I2_int_im_op_1D, (self.nz_, self.nx, self.nx))
        pass

    def interpolation_xy_vec(self, k):
        x = np.arange(-self.xymax_, 1 + self.xymax_)
        y = np.arange(-self.xymax_, 1 + self.xymax_)
        X, Y = np.meshgrid(x, y)
        xis = X - self.xp_
        yis = Y - self.yp_
        rxs = np.sqrt(xis * xis + yis * yis)
        r0s = np.int64(rxs)
        drs = rxs - r0s
        drs[((r0s + 1) >= self.rmax_)] = 0

        I0_int_re_op_1Ds = drs * self.I0_int_re[k, r0s + 1] + (1.0 - drs) * self.I0_int_re[k, r0s]

        I1_int_re_op_1Ds = drs * self.I1_int_re[k, r0s + 1] + (1.0 - drs) * self.I1_int_re[k, r0s]
        I2_int_re_op_1Ds = drs * self.I2_int_re[k, r0s + 1] + (1.0 - drs) * self.I2_int_re[k, r0s]
        I0_int_im_op_1Ds = drs * self.I0_int_im[k, r0s + 1] + (1.0 - drs) * self.I0_int_im[k, r0s]
        I1_int_im_op_1Ds = drs * self.I1_int_im[k, r0s + 1] + (1.0 - drs) * self.I1_int_im[k, r0s]
        I2_int_im_op_1Ds = drs * self.I2_int_im[k, r0s + 1] + (1.0 - drs) * self.I2_int_im[k, r0s]

        dims = np.shape(I0_int_re_op_1Ds)
        self.I0_int_re_op_1D[(k * dims[0] * dims[1]) : ((k + 1) * dims[0] * dims[1])] += np.ravel(
            I0_int_re_op_1Ds
        )
        self.I1_int_re_op_1D[(k * dims[0] * dims[1]) : ((k + 1) * dims[0] * dims[1])] += np.ravel(
            I1_int_re_op_1Ds
        )
        self.I2_int_re_op_1D[(k * dims[0] * dims[1]) : ((k + 1) * dims[0] * dims[1])] += np.ravel(
            I2_int_re_op_1Ds
        )
        self.I0_int_im_op_1D[(k * dims[0] * dims[1]) : ((k + 1) * dims[0] * dims[1])] += np.ravel(
            I0_int_im_op_1Ds
        )
        self.I1_int_im_op_1D[(k * dims[0] * dims[1]) : ((k + 1) * dims[0] * dims[1])] += np.ravel(
            I1_int_im_op_1Ds
        )
        self.I2_int_im_op_1D[(k * dims[0] * dims[1]) : ((k + 1) * dims[0] * dims[1])] += np.ravel(
            I2_int_im_op_1Ds
        )

    def theta_equal_to_alpha(self, ci, constJ, k, p, phC):
        # theta = alpha;
        sintheta = math.sin(p.alpha)
        costheta = math.cos(p.alpha)
        sqrtcostheta = math.sqrt(costheta)
        nsroot = cmath.sqrt((p.ns**2 - p.NA**2))
        ngroot = cmath.sqrt((p.ng**2 - p.NA**2))
        ts1ts2 = 4.0 * p.ni * costheta * ngroot
        tp1tp2 = ts1ts2
        tp1tp2 /= (p.ng * costheta + p.ni / p.ng * ngroot) * (
            p.ns / p.ng * ngroot + p.ng / p.ns * nsroot
        )
        ts1ts2 /= (p.ni * costheta + ngroot) * (ngroot + nsroot)
        bessel_0 = scipy.special.j0(constJ * sintheta) * sintheta * sqrtcostheta
        bessel_1 = scipy.special.j1(constJ * sintheta) * sintheta * sqrtcostheta
        if constJ != 0.0:
            bessel_2 = 2.0 * bessel_1 / (constJ * sintheta) - bessel_0
        else:
            bessel_2 = 0.0
        bessel_0 *= ts1ts2 + tp1tp2 / p.ns * nsroot
        bessel_1 *= tp1tp2 * p.ni / p.ns * sintheta
        bessel_2 *= ts1ts2 - tp1tp2 / p.ns * nsroot

        con_ti_up = p.tg0 / p.ng0 - p.tg / p.ng + p.ti0 / p.ni0 - self.zp_ / p.ns
        con_refOPD_up = p.ng0 * p.tg0 + p.ni0 * p.ti0
        ti_up = self.zp_ - self.z_[k] + p.ni * con_ti_up
        refOPD_up = -1 * ((p.ng * p.tg + p.ni * ti_up) - con_refOPD_up)
        expW = cmath.exp(
            1j * p.k0 * p.ns * self.zp_
            + 1j * phC
            + 1j * p.k0 * refOPD_up
            + 1j
            * p.k0
            * (
                (ci - self.z_[k]) * cmath.sqrt((p.ni**2 - p.NA**2))
                + self.zp_ * nsroot
                + p.tg * ngroot
                - p.tg0 * math.sqrt((p.ng0**2 - p.NA**2))
                - p.ti0 * cmath.sqrt((p.ni0**2 - p.NA**2))
            )
        )
        sum_I0 = expW * bessel_0
        sum_I1 = expW * bessel_1
        sum_I2 = expW * bessel_2
        return sum_I0, sum_I1, sum_I2

    def simpsonsRuleVec(self, ci, constJ, k, nSamples, p, phC, step, oddVal):
        ran_ = np.arange(1, int(nSamples / 2) + oddVal)
        thetas = (2 * ran_ - oddVal) * step
        sinthetas = np.sin(thetas)
        costhetas = np.cos(thetas)
        sqrtcosthetas = np.sqrt(costhetas)
        ni2sin2thetas = p.ni**2 * sinthetas**2
        nsroots = np.sqrt(p.ns**2 - ni2sin2thetas + 0j)
        ngroots = np.sqrt(p.ng**2 - ni2sin2thetas + 0j)

        ts1ts2s = 4 * p.ni * costhetas * ngroots
        tp1tp2s = ts1ts2s
        tp1tp2s = np.divide(
            tp1tp2s,
            (p.ng * costhetas + p.ni / p.ng * ngroots)
            * (p.ns / p.ng * ngroots + p.ng / p.ns * nsroots),
        )
        ts1ts2s = np.divide(ts1ts2s, (p.ni * costhetas + ngroots) * (ngroots + nsroots))

        bessel_0s = (
            2 ** (1 + oddVal) * scipy.special.j0(constJ * sinthetas) * sinthetas * sqrtcosthetas
        )
        bessel_1s = (
            2 ** (1 + oddVal) * scipy.special.j1(constJ * sinthetas) * sinthetas * sqrtcosthetas
        )
        if constJ != 0:
            bessel_2s = 2 * bessel_1s / (constJ * sinthetas) - bessel_0s
        else:
            bessel_2s = 0 * bessel_1s

        bessel_0s = np.multiply(bessel_0s, (ts1ts2s + tp1tp2s / p.ns * nsroots))
        bessel_1s = np.multiply(bessel_1s, (tp1tp2s * p.ni / p.ns * sinthetas))
        bessel_2s = np.multiply(bessel_2s, (ts1ts2s - tp1tp2s / p.ns * nsroots))

        con_ti_up = p.tg0 / p.ng0 - p.tg / p.ng + p.ti0 / p.ni0 - self.zp_ / p.ns
        con_refOPD_up = p.ng0 * p.tg0 + p.ni0 * p.ti0
        ti_up = self.zp_ - self.z_[k] + p.ni * con_ti_up
        refOPD_up = -1 * ((p.ng * p.tg + p.ni * ti_up) - con_refOPD_up)

        expWs = np.exp(
            1j * p.k0 * p.ns * self.zp_
            + 1j * phC
            + 1j * p.k0 * refOPD_up
            + 1j
            * p.k0
            * (
                (ci - self.z_[k]) * p.ni * costhetas
                + self.zp_ * nsroots
                + p.tg * ngroots
                - p.tg0 * np.sqrt(p.ng0**2 - ni2sin2thetas + 0j)
                - p.ti0 * np.sqrt(p.ni0**2 - ni2sin2thetas + 0j)
            )
        )

        sum_I0 = np.sum(expWs * bessel_0s)
        sum_I1 = np.sum(expWs * bessel_1s)
        sum_I2 = np.sum(expWs * bessel_2s)

        return sum_I0, sum_I1, sum_I2



import os
import time

import pandas as pd
from PySide6 import QtCore, QtWidgets

from piscat.GUI.BatchAnalysis.tab_analysis_protein import ProteinAnalysis_GUI
from piscat.GUI.BatchAnalysis.tab_bg_correction import BgCorrection_GUI
from piscat.GUI.BatchAnalysis.tab_localization import Localization_GUI
from piscat.GUI.BatchAnalysis.tab_plot_histogram import Histogram_GUI
from piscat.GUI.BatchAnalysis.tab_reading_data import LoadData_GUI
from piscat.GUI.BatchAnalysis.tab_trajectory import Tracking_GUI
from piscat.InputOutput import read_write_data
from piscat.InputOutput.read_write_data import load_dict_from_hdf5


class BatchAnalysis(QtWidgets.QMainWindow):
    def __init__(self, object_update_progressBar, parent=None):
        super(BatchAnalysis, self).__init__(parent)
        self.window = QtWidgets.QWidget()

        self.all_tabs = {}
        self.type_bgCorrection = None
        self.current_frame = 0
        self.object_update_progressBar = object_update_progressBar

        self.type_file = None
        self.dirName = None
        self.name_mkdir = None

        self.hyperparameters = {
            "im_size_x": None,
            "im_size_y": None,
            "start_fr": None,
            "end_fr": None,
            "image_format": None,
            "batch_size": None,
            "mode_FPN": "mFPN",
            "select_correction_axis": "Both",
            "function": None,
            "Mode_PSF_Segmentation": "BOTH",
            "min_sigma": None,
            "max_sigma": None,
            "sigma_ratio": None,
            "PSF_detection_thr": None,
            "overlap": None,
            "search_range": None,
            "memory": None,
            "outlier_frames_thr": None,
            "symmetric_PSFs_thr": None,
            "min_V_shape_width": None,
        }

        self.flags = {
            "PN": None,
            "DRA": None,
            "FPNc": None,
            "outlier_frames_filter": None,
            "Dense_Filter": None,
            "symmetric_PSFs_Filter": None,
            "FFT_flag": None,
        }

        self.history = {
            "Type_bg_correction": None,
            "bg_setting": None,
            "PSFs_Localization_setting": None,
            "PSFs_Tracking_setting": None,
        }

        self.PSFs_Particels_num = {"#PSFs": None, "#Particles": None}

        self.initUI()

        self.setWindowTitle("Batch analysis")
        self.window.setLayout(self.layout)
        self.setCentralWidget(self.window)
        self.window.show()

    def initUI(self):
        self.resize(400, 300)
        self.setWindowTitle("Batch analysis")
        self.setGeometry(1140, 100, 400, 200)

        # Tabs
        self.input_video = None
        self.batch_size = None
        self.all_tabs = {
            "Load_data": LoadData_GUI(),
            "Bg_correction": BgCorrection_GUI(),
            "PSFs_Localization": Localization_GUI(),
            "PSFs_Tracking": Tracking_GUI(self.batch_size),
            "Protein_Analysis": ProteinAnalysis_GUI(),
            "Histogram": Histogram_GUI(),
        }

        self.all_tabs["Load_data"].update_output_external.connect(self.update_read_data)
        self.all_tabs["Load_data"].update_tab_index.connect(self.update_tab)

        self.all_tabs["Bg_correction"].output_setting_Tab_bgCorrection.connect(self.update_BGc)
        self.all_tabs["Bg_correction"].update_tab_index.connect(self.update_tab)

        self.all_tabs["PSFs_Localization"].output_setting_Tab_Localization_external.connect(
            self.Update_tab_localization
        )
        self.all_tabs["PSFs_Localization"].update_tab_index.connect(self.update_tab)

        self.all_tabs["PSFs_Tracking"].output_setting_Tab_tracking.connect(
            self.Update_tab_trajectories
        )
        self.all_tabs["PSFs_Tracking"].update_tab_index.connect(self.update_tab)

        self.main_tabWidget = QtWidgets.QTabWidget()
        self.main_tabWidget.addTab(self.all_tabs["Load_data"], "Load data")
        self.main_tabWidget.addTab(self.all_tabs["Bg_correction"], "Background correction")
        self.main_tabWidget.addTab(self.all_tabs["PSFs_Localization"], "PSFs localization")
        self.main_tabWidget.addTab(self.all_tabs["PSFs_Tracking"], "PSFs tracking")
        self.main_tabWidget.addTab(self.all_tabs["Protein_Analysis"], "Protein analysis")
        self.main_tabWidget.addTab(self.all_tabs["Histogram"], "Histogram")

        self.btn_save = QtWidgets.QPushButton("Save Data_History")
        self.btn_save.setAutoDefault(False)
        self.btn_save.clicked.connect(self.save_data)
        self.btn_save.setFixedWidth(120)
        self.btn_save.setFixedHeight(20)

        self.btn_load_csv = QtWidgets.QPushButton("Load CSV")
        self.btn_load_csv.setAutoDefault(False)
        self.btn_load_csv.clicked.connect(self.load_csv_data)
        self.btn_load_csv.setFixedWidth(100)
        self.btn_load_csv.setFixedHeight(20)

        self.btn_load_hdf5 = QtWidgets.QPushButton("Load hdf5")
        self.btn_load_hdf5.setAutoDefault(False)
        self.btn_load_hdf5.clicked.connect(self.load_hdf5_data)
        self.btn_load_hdf5.setFixedWidth(120)
        self.btn_load_hdf5.setFixedHeight(20)

        self.statusbar = self.statusBar()
        self.statusBar().addPermanentWidget(self.btn_save)
        self.statusBar().addPermanentWidget(self.btn_load_csv)
        self.statusBar().addPermanentWidget(self.btn_load_hdf5)

        self.main_tabWidget.setCurrentIndex(0)

        self.layout = QtWidgets.QVBoxLayout()
        self.layout.addWidget(self.main_tabWidget)
        self.layout.setContentsMargins(0, 0, 0, 0)

    def __del__(self):
        print("Destructor called, Employee deleted.")

    @QtCore.Slot(int)
    def update_tab(self, idx_):
        if idx_ == 4:
            self.all_tabs["Protein_Analysis"].updata_input_setting(
                [self.hyperparameters, self.flags]
            )
            self.main_tabWidget.setCurrentIndex(idx_)
        else:
            self.main_tabWidget.setCurrentIndex(idx_)

    @QtCore.Slot(int)
    def update_read_data(self, data_in):
        self.hyperparameters["im_size_x"] = data_in["img_width"]
        self.hyperparameters["im_size_y"] = data_in["img_height"]
        self.hyperparameters["width_size_s"] = data_in["width_size_s"]
        self.hyperparameters["width_size_e"] = data_in["width_size_e"]
        self.hyperparameters["height_size_s"] = data_in["height_size_s"]
        self.hyperparameters["height_size_e"] = data_in["height_size_e"]
        self.hyperparameters["start_fr"] = data_in["s_frame"]
        self.hyperparameters["end_fr"] = data_in["e_frame"]
        self.hyperparameters["frame_stride"] = data_in["frame_stride"]
        self.hyperparameters["image_format"] = data_in["image_type"]
        self.hyperparameters["type_file"] = data_in["type_file"]
        self.hyperparameters["path"] = data_in["path"]
        self.hyperparameters["name_mkdir"] = data_in["name_mkdir"]

    @QtCore.Slot()
    def update_BGc(self, data_in):
        for key in data_in:
            if key == "Batch_size (frames)":
                self.hyperparameters["batch_size"] = data_in[key]
            elif key == "FPNc_axis":
                self.hyperparameters["select_correction_axis"] = data_in[key]
            elif key == "Power_Normalization":
                self.flags["PN"] = data_in["Power_Normalization"]
            elif key == "FPNc":
                self.flags[key] = data_in[key]
            elif key == "filter_hotPixels":
                self.flags[key] = data_in[key]
            elif key == "type_BGc":
                if data_in["type_BGc"] == "DRA":
                    self.flags["DRA"] = True
                else:
                    self.flags["DRA"] = False
            else:
                self.hyperparameters[key] = data_in[key]

        self.flags["FFT_flag"] = True

    @QtCore.Slot()
    def Update_tab_localization(self, data_in):
        for key in data_in:
            if key == "mode":
                self.hyperparameters["Mode_PSF_Segmentation"] = data_in[key]
            elif key == "threshold_min":
                self.hyperparameters["PSF_detection_thr"] = data_in[key]
            elif key == "Asymmetry_PSFs_filtering_threshold":
                self.hyperparameters["symmetric_PSFs_thr"] = data_in[key]
            elif key == "Outlier_frames_filter_max_number_PSFs":
                self.hyperparameters["outlier_frames_thr"] = data_in[key]
            elif key == "Flag_fine_localization":
                self.flags["Fine_localization"] = data_in[key]
            elif key == "Flag_fit_2DGaussian":
                self.flags["Flag_fit_2DGaussian"] = data_in[key]
            elif (
                key == "Side_lobes_Filter"
                or key == "Dense_Filter"
                or key == "outlier_frames_filter"
                or key == "symmetric_PSFs_Filter"
            ):
                self.flags[key] = data_in[key]
            else:
                self.hyperparameters[key] = data_in[key]

    @QtCore.Slot()
    def Update_tab_trajectories(self, data_in):
        self.hyperparameters["search_range"] = data_in["Neighborhood_size (px)"]
        self.hyperparameters["memory"] = data_in["Memory (frame)"]
        self.hyperparameters["min_V_shape_width"] = data_in["Minimum_temporal_length (frame)"]

    def save_data(self):
        self.file_path = False
        self.file_path = QtWidgets.QFileDialog.getExistingDirectory(
            self, "Select Folder", os.path.expanduser("~"), QtWidgets.QFileDialog.ShowDirsOnly
        )
        timestr = time.strftime("%Y%m%d-%H%M%S")
        name_mkdir = timestr
        try:
            dr_mk = os.path.join(self.file_path, name_mkdir)
            os.mkdir(dr_mk)
            print("Directory ", name_mkdir, " Created ")
        except FileExistsError:
            dr_mk = os.path.join(self.file_path, name_mkdir)
            print("Directory ", name_mkdir, " already exists")

        self.file_path = dr_mk

        if self.file_path:
            if self.df_PSFs is not None:
                read_write_data.save_df2csv(self.df_PSFs, path=self.file_path, name="df_PSFs")

            fout = os.path.join(self.file_path, "history.txt")
            fo = open(fout, "w")
            for k, v in self.history.items():
                fo.write(str(k) + " >>> " + str(v) + "\n\n")

            for k, v in self.PSFs_Particels_num.items():
                fo.write(str(k) + " >>> " + str(v) + "\n\n")
            fo.close()

            if self.trajectories is not None:
                read_write_data.save_mat(
                    data=self.trajectories, path=self.file_path, name="all_trajectories"
                )
                read_write_data.save_list_to_hdf5(
                    list_data=self.trajectories, path=self.file_path, name="histData"
                )

    def load_hdf5_data(self):
        if self.bgCorrectedVideo is not None:
            try:
                fileName = QtWidgets.QFileDialog().getOpenFileName()
                filePath = str(fileName[0])
                data_dic = load_dict_from_hdf5(filePath)
                self.Update_tab_trajectories(data_dic)
            except:
                print("HDF5 can not load!")
        elif self.bgCorrectedVideo is None:
            self.msg_box1 = QtWidgets.QMessageBox()
            self.msg_box1.setWindowTitle("Warning!")
            self.msg_box1.setText("Update background tab!")
            self.msg_box1.exec_()

    def load_csv_data(self):
        if self.bgCorrectedVideo is not None:
            try:
                fileName = QtWidgets.QFileDialog().getOpenFileName()
                filePath = str(fileName[0])
                df_PSFs = pd.read_csv(filePath)
                self.Update_tab_localization(df_PSFs)
            except:
                print("CSV can not load!")
        elif self.bgCorrectedVideo is None:
            self.msg_box1 = QtWidgets.QMessageBox()
            self.msg_box1.setWindowTitle("Warning!")
            self.msg_box1.setText("Update background tab!")
            self.msg_box1.exec_()
